# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

import math

from copy import copy as _copy
from logging import DEBUG as _DEBUG
from logging import getLogger

from .common import NamedInt as _NamedInt
from .common import NamedInts as _NamedInts
from .common import bytes2int as _bytes2int
from .common import int2bytes as _int2bytes

_log = getLogger(__name__)
del getLogger

#
#
#

KIND = _NamedInts(toggle=0x01, choice=0x02, range=0x04, map_choice=0x0A, multiple_toggle=0x10)


class Setting(object):
    """A setting descriptor.
    Needs to be instantiated for each specific device."""
    __slots__ = (
        'name', 'label', 'description', 'kind', 'device_kind', 'feature', 'persist', '_rw', '_validator', '_device', '_value'
    )

    def __init__(
        self,
        name,
        rw,
        validator,
        kind=None,
        label=None,
        description=None,
        device_kind=None,
        feature=None,
        persist=True,
        **kwargs
    ):
        assert name
        self.name = name
        self.label = label or name
        self.description = description
        self.device_kind = device_kind
        self.feature = feature
        self.persist = persist

        self._rw = rw
        self._validator = validator

        assert kind is None or kind & validator.kind != 0
        self.kind = kind or validator.kind

    def __call__(self, device):
        assert not hasattr(self, '_value')
        # combined keyboards and touchpads (e.g., K400) break this assertion so don't use it
        # assert self.device_kind is None or device.kind in self.device_kind
        p = device.protocol
        if p == 1.0:
            # HID++ 1.0 devices do not support features
            assert self._rw.kind == RegisterRW.kind
        elif p >= 2.0:
            # HID++ 2.0 devices do not support registers
            assert self._rw.kind == FeatureRW.kind

        o = _copy(self)
        o._value = None
        o._device = device
        return o

    @property
    def choices(self):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')

        return self._validator.choices if self._validator.kind & KIND.choice else None

    @property
    def range(self):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')

        if self._validator.kind == KIND.range:
            return (self._validator.min_value, self._validator.max_value)

    def read(self, cached=True):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')

        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: settings read %r from %s', self.name, self._value, self._device)

        if self.persist and self._value is None and self._device.persister:
            # We haven't read a value from the device yet,
            # maybe we have something in the configuration.
            self._value = self._device.persister.get(self.name)

        if cached and self._value is not None:
            if self.persist and self._device.persister and self.name not in self._device.persister:
                # If this is a new device (or a new setting for an old device),
                # make sure to save its current value for the next time.
                self._device.persister[self.name] = self._value
            return self._value

        if self._device.online:
            reply = self._rw.read(self._device)
            if reply:
                self._value = self._validator.validate_read(reply)
            if self.persist and self._device.persister and self.name not in self._device.persister:
                # Don't update the persister if it already has a value,
                # otherwise the first read might overwrite the value we wanted.
                self._device.persister[self.name] = self._value
            return self._value

    def write(self, value):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')
        assert value is not None

        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: settings write %r to %s', self.name, value, self._device)

        if self._device.online:
            # Remember the value we're trying to set, even if the write fails.
            # This way even if the device is offline or some other error occurs,
            # the last value we've tried to write is remembered in the configuration.
            self._value = value
            if self.persist and self._device.persister:
                self._device.persister[self.name] = value

            current_value = None
            if self._validator.needs_current_value:
                # the validator needs the current value, possibly to merge flag values
                current_value = self._rw.read(self._device)

            data_bytes = self._validator.prepare_write(value, current_value)
            if data_bytes is not None:
                if _log.isEnabledFor(_DEBUG):
                    _log.debug('%s: settings prepare write(%s) => %r', self.name, value, data_bytes)

                reply = self._rw.write(self._device, data_bytes)
                if not reply:
                    # tell whomever is calling that the write failed
                    return None

            return value

    def apply(self):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')

        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: apply %s (%s)', self.name, self._value, self._device)

        value = self.read(self.persist)  # Don't use persisted value if setting doesn't persist
        if self.persist and value is not None:  # If setting doesn't persist no need to write value just read
            self.write(value)

    def __str__(self):
        if hasattr(self, '_value'):
            assert hasattr(self, '_device')
            return '<Setting([%s:%s] %s:%s=%s)>' % (
                self._rw.kind, self._validator.kind, self._device.codename, self.name, self._value
            )
        return '<Setting([%s:%s] %s)>' % (self._rw.kind, self._validator.kind, self.name)

    __unicode__ = __repr__ = __str__


class Settings(Setting):
    """A setting descriptor for multiple choices, being a map from keys to values.
    Needs to be instantiated for each specific device."""
    def read(self, cached=True):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')

        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: settings read %r from %s', self.name, self._value, self._device)

        if self.persist and self._value is None and getattr(self._device, 'persister', None):
            # We haven't read a value from the device yet,
            # maybe we have something in the configuration.
            self._value = self._device.persister.get(self.name)

        if cached and self._value is not None:
            if self.persist and getattr(self._device, 'persister', None) and self.name not in self._device.persister:
                # If this is a new device (or a new setting for an old device),
                # make sure to save its current value for the next time.
                self._device.persister[self.name] = self._value
            return self._value

        if self._device.online:
            reply_map = {}
            for key in self._validator.choices:
                reply = self._rw.read(self._device, key)
                if reply:
                    # keys are ints, because that is what the device uses,
                    # encoded into strings because JSON requires strings as keys
                    reply_map[str(int(key))] = self._validator.validate_read(reply, key)
            self._value = reply_map
            if self.persist and getattr(self._device, 'persister', None) and self.name not in self._device.persister:
                # Don't update the persister if it already has a value,
                # otherwise the first read might overwrite the value we wanted.
                self._device.persister[self.name] = self._value
            return self._value

    def read_key(self, key, cached=True):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')
        assert key is not None

        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: settings read %r key %r from %s', self.name, self._value, key, self._device)

        if self._value is None and getattr(self._device, 'persister', None):
            self._value = self._device.persister.get(self.name)

        if cached and self._value is not None:
            if self.persist and getattr(self._device, 'persister', None) and self.name not in self._device.persister:
                self._device.persister[self.name] = self._value
            return self._value[str(int(key))]

        if self._device.online:
            reply = self._rw.read(self._device, key)
            if reply:
                self._value[str(int(key))] = self._validator.validate_read(reply, key)
            if self.persist and getattr(self._device, 'persister', None) and self.name not in self._device.persister:
                self._device.persister[self.name] = self._value
            return self._value[str(int(key))]

    def write(self, map):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')
        assert map is not None

        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: settings write %r to %s', self.name, map, self._device)

        if self._device.online:
            # Remember the value we're trying to set, even if the write fails.
            # This way even if the device is offline or some other error occurs,
            # the last value we've tried to write is remembered in the configuration.
            self._value = map
            if self.persist and self._device.persister:
                self._device.persister[self.name] = map

            for key, value in map.items():
                data_bytes = self._validator.prepare_write(int(key), value)
                if data_bytes is not None:
                    if _log.isEnabledFor(_DEBUG):
                        _log.debug('%s: settings prepare map write(%s,%s) => %r', self.name, key, value, data_bytes)
                    reply = self._rw.write(self._device, int(key), data_bytes)
                    if not reply:
                        return None

            return map

    def write_key_value(self, key, value):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')
        assert key is not None
        assert value is not None

        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: settings write key %r value %r to %s', self.name, key, value, self._device)

        if self._device.online:
            # Remember the value we're trying to set, even if the write fails.
            # This way even if the device is offline or some other error occurs,
            # the last value we've tried to write is remembered in the configuration.
            self._value[str(key)] = value
            if self.persist and self._device.persister:
                self._device.persister[self.name] = self._value

            data_bytes = self._validator.prepare_write(int(key), value)
            if data_bytes is not None:
                if _log.isEnabledFor(_DEBUG):
                    _log.debug('%s: settings prepare key value write(%s,%s) => %r', self.name, key, value, data_bytes)
                reply = self._rw.write(self._device, int(key), data_bytes)
                if not reply:
                    # tell whomever is calling that the write failed
                    return None

            return value


class BitFieldSetting(Setting):
    """A setting descriptor for a set of choices represented by one bit each, being a map from options to booleans.
    Needs to be instantiated for each specific device."""
    def read(self, cached=True):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')

        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: settings read %r from %s', self.name, self._value, self._device)

        if self._value is None and getattr(self._device, 'persister', None):
            # We haven't read a value from the device yet,
            # maybe we have something in the configuration.
            self._value = self._device.persister.get(self.name)

        if cached and self._value is not None:
            if self.persist and getattr(self._device, 'persister', None) and self.name not in self._device.persister:
                # If this is a new device (or a new setting for an old device),
                # make sure to save its current value for the next time.
                self._device.persister[self.name] = self._value
            return self._value

        if self._device.online:
            reply_map = {}
            reply = self._rw.read(self._device)
            if reply:
                # keys are ints, because that is what the device uses,
                # encoded into strings because JSON requires strings as keys
                reply_map = self._validator.validate_read(reply)
            self._value = reply_map
            if self.persist and getattr(self._device, 'persister', None) and self.name not in self._device.persister:
                # Don't update the persister if it already has a value,
                # otherwise the first read might overwrite the value we wanted.
                self._device.persister[self.name] = self._value
            return self._value

    def read_key(self, key, cached=True):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')
        assert key is not None

        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: settings read %r key %r from %s', self.name, self._value, key, self._device)

        if self._value is None and getattr(self._device, 'persister', None):
            self._value = self._device.persister.get(self.name)

        if cached and self._value is not None:
            if self.persist and getattr(self._device, 'persister', None) and self.name not in self._device.persister:
                self._device.persister[self.name] = self._value
            return self._value[str(int(key))]

        if self._device.online:
            reply = self._rw.read(self._device, key)
            if reply:
                self._value = self._validator.validate_read(reply)
            if self.persist and getattr(self._device, 'persister', None) and self.name not in self._device.persister:
                self._device.persister[self.name] = self._value
            return self._value[str(int(key))]

    def write(self, map):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')
        assert map is not None

        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: settings write %r to %s', self.name, map, self._device)

        if self._device.online:
            # Remember the value we're trying to set, even if the write fails.
            # This way even if the device is offline or some other error occurs,
            # the last value we've tried to write is remembered in the configuration.
            self._value = map
            if self.persist and self._device.persister:
                self._device.persister[self.name] = map
            data_bytes = self._validator.prepare_write(self._value)
            if data_bytes is not None:
                if _log.isEnabledFor(_DEBUG):
                    _log.debug('%s: settings prepare map write(%s) => %r', self.name, self._value, data_bytes)
                reply = self._rw.write(self._device, data_bytes)
                if not reply:
                    return None
            return map

    def write_key_value(self, key, value):
        assert hasattr(self, '_value')
        assert hasattr(self, '_device')
        assert key is not None
        assert value is not None

        if _log.isEnabledFor(_DEBUG):
            _log.debug('%s: settings write key %r value %r to %s', self.name, key, value, self._device)

        if self._device.online:
            # Remember the value we're trying to set, even if the write fails.
            # This way even if the device is offline or some other error occurs,
            # the last value we've tried to write is remembered in the configuration.
            value = bool(value)
            self._value[str(key)] = value
            if self.persist and self._device.persister:
                self._device.persister[self.name] = self._value

            data_bytes = self._validator.prepare_write(self._value)
            if data_bytes is not None:
                if _log.isEnabledFor(_DEBUG):
                    _log.debug('%s: settings prepare key value write(%s,%s) => %r', self.name, key, str(value), data_bytes)
                reply = self._rw.write(self._device, data_bytes)
                if not reply:
                    # tell whomever is calling that the write failed
                    return None

            return value


#
# read/write low-level operators
#


class RegisterRW(object):
    __slots__ = ('register', )

    kind = _NamedInt(0x01, 'register')

    def __init__(self, register):
        assert isinstance(register, int)
        self.register = register

    def read(self, device):
        return device.read_register(self.register)

    def write(self, device, data_bytes):
        return device.write_register(self.register, data_bytes)


class FeatureRW(object):
    __slots__ = ('feature', 'read_fnid', 'write_fnid', 'no_reply')

    kind = _NamedInt(0x02, 'feature')
    default_read_fnid = 0x00
    default_write_fnid = 0x10

    def __init__(self, feature, read_fnid=default_read_fnid, write_fnid=default_write_fnid, no_reply=False, **kwargs):
        assert isinstance(feature, _NamedInt)
        self.feature = feature
        self.read_fnid = read_fnid
        self.write_fnid = write_fnid
        self.no_reply = no_reply

    def read(self, device):
        assert self.feature is not None
        return device.feature_request(self.feature, self.read_fnid)

    def write(self, device, data_bytes):
        assert self.feature is not None
        reply = device.feature_request(self.feature, self.write_fnid, data_bytes, no_reply=self.no_reply)
        return reply if not self.no_reply else True


class FeatureRWMap(FeatureRW):
    kind = _NamedInt(0x02, 'feature')
    default_read_fnid = 0x00
    default_write_fnid = 0x10
    default_key_bytes_count = 1

    def __init__(
        self,
        feature,
        read_fnid=default_read_fnid,
        write_fnid=default_write_fnid,
        key_bytes_count=default_key_bytes_count,
        no_reply=False,
        **_ignore
    ):
        assert isinstance(feature, _NamedInt)
        self.feature = feature
        self.read_fnid = read_fnid
        self.write_fnid = write_fnid
        self.key_bytes_count = key_bytes_count
        self.no_reply = no_reply

    def read(self, device, key):
        assert self.feature is not None
        key_bytes = _int2bytes(key, self.key_bytes_count)
        return device.feature_request(self.feature, self.read_fnid, key_bytes)

    def write(self, device, key, data_bytes):
        assert self.feature is not None
        key_bytes = _int2bytes(key, self.key_bytes_count)
        reply = device.feature_request(self.feature, self.write_fnid, key_bytes, data_bytes, no_reply=self.no_reply)
        return reply if not self.no_reply else True


#
# value validators
# handle the conversion from read bytes, to setting value, and back
#


class BooleanValidator(object):
    __slots__ = ('true_value', 'false_value', 'mask', 'needs_current_value')

    kind = KIND.toggle
    default_true = 0x01
    default_false = 0x00
    # mask specifies all the affected bits in the value
    default_mask = 0xFF

    def __init__(self, true_value=default_true, false_value=default_false, mask=default_mask, **kwargs):
        if isinstance(true_value, int):
            assert isinstance(false_value, int)
            if mask is None:
                mask = self.default_mask
            else:
                assert isinstance(mask, int)
            assert true_value & false_value == 0
            assert true_value & mask == true_value
            assert false_value & mask == false_value
            self.needs_current_value = (mask != self.default_mask)
        elif isinstance(true_value, bytes):
            if false_value is None or false_value == self.default_false:
                false_value = b'\x00' * len(true_value)
            else:
                assert isinstance(false_value, bytes)
            if mask is None or mask == self.default_mask:
                mask = b'\xFF' * len(true_value)
            else:
                assert isinstance(mask, bytes)
            assert len(mask) == len(true_value) == len(false_value)
            tv = _bytes2int(true_value)
            fv = _bytes2int(false_value)
            mv = _bytes2int(mask)
            assert tv != fv  # true and false might be something other than bit values
            assert tv & mv == tv
            assert fv & mv == fv
            self.needs_current_value = any(m != b'\xFF' for m in mask)
        else:
            raise Exception("invalid mask '%r', type %s" % (mask, type(mask)))

        self.true_value = true_value
        self.false_value = false_value
        self.mask = mask

    def validate_read(self, reply_bytes):
        if isinstance(self.mask, int):
            reply_value = ord(reply_bytes[:1]) & self.mask
            if _log.isEnabledFor(_DEBUG):
                _log.debug('BooleanValidator: validate read %r => %02X', reply_bytes, reply_value)
            if reply_value == self.true_value:
                return True
            if reply_value == self.false_value:
                return False
            _log.warn(
                'BooleanValidator: reply %02X mismatched %02X/%02X/%02X', reply_value, self.true_value, self.false_value,
                self.mask
            )
            return False

        count = len(self.mask)
        mask = _bytes2int(self.mask)
        reply_value = _bytes2int(reply_bytes[:count]) & mask

        true_value = _bytes2int(self.true_value)
        if reply_value == true_value:
            return True

        false_value = _bytes2int(self.false_value)
        if reply_value == false_value:
            return False

        _log.warn('BooleanValidator: reply %r mismatched %r/%r/%r', reply_bytes, self.true_value, self.false_value, self.mask)
        return False

    def prepare_write(self, new_value, current_value=None):
        if new_value is None:
            new_value = False
        else:
            assert isinstance(new_value, bool)

        to_write = self.true_value if new_value else self.false_value

        if isinstance(self.mask, int):
            if current_value is not None and self.needs_current_value:
                to_write |= ord(current_value[:1]) & (0xFF ^ self.mask)
            if current_value is not None and to_write == ord(current_value[:1]):
                return None
        else:
            to_write = bytearray(to_write)
            count = len(self.mask)
            for i in range(0, count):
                b = ord(to_write[i:i + 1])
                m = ord(self.mask[i:i + 1])
                assert b & m == b
                # b &= m
                if current_value is not None and self.needs_current_value:
                    b |= ord(current_value[i:i + 1]) & (0xFF ^ m)
                to_write[i] = b
            to_write = bytes(to_write)

            if current_value is not None and to_write == current_value[:len(to_write)]:
                return None

        if _log.isEnabledFor(_DEBUG):
            _log.debug('BooleanValidator: prepare_write(%s, %s) => %r', new_value, current_value, to_write)

        return to_write


class BitFieldValidator(object):
    __slots__ = ('byte_count', 'options')

    kind = KIND.multiple_toggle

    def __init__(self, options, byte_count=None, **kwargs):
        assert (isinstance(options, list))
        self.options = options
        self.byte_count = (max(x.bit_length() for x in options) + 7) // 8
        if byte_count:
            assert (isinstance(byte_count, int) and byte_count >= self.byte_count)
            self.byte_count = byte_count

    def validate_read(self, reply_bytes):
        r = _bytes2int(reply_bytes[:self.byte_count])
        value = {str(int(k)): False for k in self.options}
        m = 1
        for _ in range(8 * self.byte_count):
            if m in self.options:
                value[str(int(m))] = bool(r & m)
            m <<= 1
        return value

    def prepare_write(self, new_value):
        assert (isinstance(new_value, dict))
        w = 0
        for k, v in new_value.items():
            if v:
                w |= int(k)
        return _int2bytes(w, self.byte_count)


class ChoicesValidator(object):
    kind = KIND.choice
    """Translates between NamedInts and a byte sequence.
    :param choices: a list of NamedInts
    :param bytes_count: the size of the derived byte sequence. If None, it
    will be calculated from the choices."""
    def __init__(self, choices, bytes_count=None, read_skip_bytes_count=None, write_prefix_bytes=b'', **_ignore):
        assert choices is not None
        assert isinstance(choices, _NamedInts)
        assert len(choices) > 1
        self.choices = choices
        self.needs_current_value = False

        max_bits = max(x.bit_length() for x in choices)
        self._bytes_count = (max_bits // 8) + (1 if max_bits % 8 else 0)
        if bytes_count:
            assert self._bytes_count <= bytes_count
            self._bytes_count = bytes_count
        assert self._bytes_count < 8
        self._read_skip_bytes_count = read_skip_bytes_count if read_skip_bytes_count else 0
        self._write_prefix_bytes = write_prefix_bytes if write_prefix_bytes else b''
        assert self._bytes_count + self._read_skip_bytes_count <= 14
        assert self._bytes_count + len(self._write_prefix_bytes) <= 14

    def validate_read(self, reply_bytes):
        reply_value = _bytes2int(reply_bytes[self._read_skip_bytes_count:self._read_skip_bytes_count + self._bytes_count])
        valid_value = self.choices[reply_value]
        assert valid_value is not None, '%s: failed to validate read value %02X' % (self.__class__.__name__, reply_value)
        return valid_value

    def prepare_write(self, new_value, current_value=None):
        if new_value is None:
            choice = self.choices[:][0]
        else:
            if isinstance(new_value, int):
                choice = self.choices[new_value]
            elif int(new_value) in self.choices:
                choice = self.choices[int(new_value)]
            elif new_value in self.choices:
                choice = self.choices[new_value]
            else:
                raise ValueError(new_value)

        if choice is None:
            raise ValueError('invalid choice %r' % new_value)
        assert isinstance(choice, _NamedInt)
        return self._write_prefix_bytes + choice.bytes(self._bytes_count)


class ChoicesMapValidator(ChoicesValidator):
    kind = KIND.map_choice

    def __init__(
        self,
        choices_map,
        key_bytes_count=None,
        bytes_count=None,
        read_skip_bytes_count=0,
        write_prefix_bytes=b'',
        extra_default=None,
        **kwargs
    ):
        assert choices_map is not None
        assert isinstance(choices_map, dict)
        max_key_bits = 0
        max_value_bits = 0
        for key, choices in choices_map.items():
            assert isinstance(key, _NamedInt)
            assert isinstance(choices, list)
            max_key_bits = max(max_key_bits, key.bit_length())
            for key_value in choices:
                assert isinstance(key_value, _NamedInt)
                max_value_bits = max(max_value_bits, key_value.bit_length())
        self._key_bytes_count = (max_key_bits + 7) // 8
        if key_bytes_count:
            assert self._key_bytes_count <= key_bytes_count
            self._key_bytes_count = key_bytes_count
        self._bytes_count = (max_value_bits + 7) // 8
        if bytes_count:
            assert self._bytes_count <= bytes_count
            self._bytes_count = bytes_count
        self.choices = choices_map
        self.needs_current_value = False
        self.extra_default = extra_default
        self._read_skip_bytes_count = read_skip_bytes_count if read_skip_bytes_count else 0
        self._write_prefix_bytes = write_prefix_bytes if write_prefix_bytes else b''
        assert self._bytes_count + self._read_skip_bytes_count + self._key_bytes_count <= 14
        assert self._bytes_count + len(self._write_prefix_bytes) + self._key_bytes_count <= 14

    def validate_read(self, reply_bytes, key):
        start = self._key_bytes_count + self._read_skip_bytes_count
        end = start + self._bytes_count
        reply_value = _bytes2int(reply_bytes[start:end])
        # reprogrammable keys starts out as 0, which is not a choice, so don't use assert here
        if self.extra_default is not None and self.extra_default == reply_value:
            return int(self.choices[key][0])
        assert reply_value in self.choices[
            key], '%s: failed to validate read value %02X' % (self.__class__.__name__, reply_value)
        return reply_value

    def prepare_write(self, key, new_value):
        choices = self.choices[key]
        if new_value not in choices and new_value != self.extra_default:
            raise ValueError('invalid choice %r' % new_value)
        return self._write_prefix_bytes + new_value.to_bytes(self._bytes_count, 'big')


class RangeValidator(object):
    __slots__ = ('min_value', 'max_value', 'flag', '_bytes_count', 'needs_current_value')

    kind = KIND.range
    """Translates between integers and a byte sequence.
    :param min_value: minimum accepted value (inclusive)
    :param max_value: maximum accepted value (inclusive)
    :param bytes_count: the size of the derived byte sequence. If None, it
    will be calculated from the range."""
    def __init__(self, min_value, max_value, bytes_count=None, **kwargs):
        assert max_value > min_value
        self.min_value = min_value
        self.max_value = max_value
        self.needs_current_value = False

        self._bytes_count = math.ceil(math.log(max_value + 1, 256))
        if bytes_count:
            assert self._bytes_count <= bytes_count
            self._bytes_count = bytes_count
        assert self._bytes_count < 8

    def validate_read(self, reply_bytes):
        reply_value = _bytes2int(reply_bytes[:self._bytes_count])
        assert reply_value >= self.min_value, '%s: failed to validate read value %02X' % (self.__class__.__name__, reply_value)
        assert reply_value <= self.max_value, '%s: failed to validate read value %02X' % (self.__class__.__name__, reply_value)
        return reply_value

    def prepare_write(self, new_value, current_value=None):
        if new_value < self.min_value or new_value > self.max_value:
            raise ValueError('invalid choice %r' % new_value)
        return _int2bytes(new_value, self._bytes_count)



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

from collections import namedtuple

from .common import NamedInts as _NamedInts
from .hidpp10 import DEVICE_KIND as _DK
from .hidpp10 import REGISTERS as _R
from .settings_templates import FeatureSettings as _FS
from .settings_templates import RegisterSettings as _RS

#
#
#

_DeviceDescriptor = namedtuple(
    '_DeviceDescriptor', ('name', 'kind', 'wpid', 'codename', 'protocol', 'registers', 'settings', 'persister')
)
del namedtuple

DEVICES = {}


def _D(name, codename=None, kind=None, wpid=None, protocol=None, registers=None, settings=None, persister=None):
    assert name

    if kind is None:
        kind = (
            _DK.mouse if 'Mouse' in name else _DK.keyboard if 'Keyboard' in name else _DK.numpad
            if 'Number Pad' in name else _DK.touchpad if 'Touchpad' in name else _DK.trackball if 'Trackball' in name else None
        )
    assert kind is not None, 'descriptor for %s does not have kind set' % name

    # heuristic: the codename is the last word in the device name
    if codename is None and ' ' in name:
        codename = name.split(' ')[-1]
    assert codename is not None, 'descriptor for %s does not have codename set' % name

    if protocol is not None:
        # ? 2.0 devices should not have any registers
        _kind = lambda s: s._rw.kind if hasattr(s, '_rw') else s._rw_kind
        if protocol < 2.0:
            assert settings is None or all(_kind(s) == 1 for s in settings)
        else:
            assert registers is None
            assert settings is None or all(_kind(s) == 2 for s in settings)

        if wpid:
            for w in wpid if isinstance(wpid, tuple) else (wpid, ):
                if protocol > 1.0:
                    assert w[0:1] == '4', '%s has protocol %0.1f, wpid %s' % (name, protocol, w)
                else:
                    if w[0:1] == '1':
                        assert kind == _DK.mouse, '%s has protocol %0.1f, wpid %s' % (name, protocol, w)
                    elif w[0:1] == '2':
                        assert kind in (_DK.keyboard, _DK.numpad), '%s has protocol %0.1f, wpid %s' % (name, protocol, w)

    device_descriptor = _DeviceDescriptor(
        name=name,
        kind=kind,
        wpid=wpid,
        codename=codename,
        protocol=protocol,
        registers=registers,
        settings=settings,
        persister=persister
    )

    assert codename not in DEVICES, 'duplicate codename in device descriptors: %s' % (DEVICES[codename], )
    DEVICES[codename] = device_descriptor

    if wpid:
        if not isinstance(wpid, tuple):
            wpid = (wpid, )

        for w in wpid:
            assert w not in DEVICES, 'duplicate wpid in device descriptors: %s' % (DEVICES[w], )
            DEVICES[w] = device_descriptor


#
#
#

_PERFORMANCE_MX_DPIS = _NamedInts.range(0x81, 0x8F, lambda x: str((x - 0x80) * 100))

#
#
#

# Some HID++1.0 registers and HID++2.0 features can be discovered at run-time,
# so they are not specified here.
#
# For known registers, however, please do specify them here -- avoids
# unnecessary communication with the device and makes it easier to make certain
# decisions when querying the device's state.
#
# Specify a negative value to blacklist a certain register for a device.
#
# Usually, state registers (battery, leds, some features, etc) are only used by
# HID++ 1.0 devices, while HID++ 2.0 devices use features for the same
# functionalities. This is a rule that's been discovered by trial-and-error,
# so it may change in the future.

# Well-known registers (in hex):
#  * 00 - notification flags (all devices)
#    01 - mice: smooth scrolling
#    07 - battery status
#    09 - keyboards: FN swap (if it has the FN key)
#    0D - battery charge
#       a device may have either the 07 or 0D register available;
#       no known device uses both
#    51 - leds
#    63 - mice: DPI
#  * F1 - firmware info
# Some registers appear to be universally supported, no matter the HID++ version
# (marked with *). The rest may or may not be supported, and their values may or
# may not mean the same thing across different devices.

# The 'codename' and 'kind' fields are usually guessed from the device name,
# but in some cases (like the Logitech Cube) that heuristic fails and they have
# to be specified.
#
# The 'protocol' and 'wpid' fields are optional (they can be discovered at
# runtime), but specifying them here speeds up device discovery and reduces the
# USB traffic Solaar has to do to fully identify peripherals.
# Same goes for HID++ 2.0 feature settings (like _feature_fn_swap).
#
# The 'registers' field indicates read-only registers, specifying a state. These
# are valid (AFAIK) only to HID++ 1.0 devices.
# The 'settings' field indicates a read/write register; based on them Solaar
# generates, at runtime, the settings controls in the device panel. HID++ 1.0
# devices may only have register-based settings; HID++ 2.0 devices may only have
# feature-based settings.

# Keyboards

_D('Wireless Keyboard K230', protocol=2.0, wpid='400D')
_D('Wireless Keyboard K270(unifying)', protocol=2.0, wpid='4003')
_D(
    'Wireless Keyboard MK270',
    protocol=2.0,
    wpid='4023',
    settings=[_FS.fn_swap()],
)
_D(
    'Wireless Keyboard K270',
    protocol=1.0,
    registers=(_R.battery_status, ),
)
_D(
    'Wireless Keyboard MK300',
    protocol=1.0,
    wpid='8521',
    registers=(_R.battery_status, ),
)

_D(
    'Wireless Keyboard MK320',
    protocol=1.0,
    wpid='200F',
    registers=(_R.battery_status, ),
)
_D('Wireless Keyboard MK330')
_D(
    'Wireless Compact Keyboard K340',
    protocol=1.0,
    wpid='2007',
    registers=(_R.battery_status, ),
)
_D(
    'Wireless Wave Keyboard K350',
    protocol=1.0,
    wpid='200A',
    registers=(_R.battery_status, ),
)
_D(
    'Wireless Keyboard K360',
    protocol=2.0,
    wpid='4004',
    settings=[_FS.fn_swap()],
)
_D(
    'Wireless Keyboard K375s',
    protocol=2.0,
    wpid='4061',
    settings=[_FS.k375s_fn_swap()],
)
_D(
    'Wireless Touch Keyboard K400',
    protocol=2.0,
    wpid=('400E', '4024'),
    settings=[_FS.fn_swap()],
)
_D(
    'Wireless Touch Keyboard K400 Plus',
    codename='K400 Plus',
    protocol=2.0,
    wpid='404D',
    settings=[
        _FS.new_fn_swap(),
        _FS.reprogrammable_keys(),
        _FS.disable_keyboard_keys(),
    ],
)
_D(
    'Wireless Keyboard K520',
    protocol=1.0,
    wpid='2011',
    registers=(_R.battery_status, ),
    settings=[
        _RS.fn_swap(),
    ],
)
_D(
    'Number Pad N545',
    protocol=1.0,
    wpid='2006',
    registers=(_R.battery_status, ),
)
_D('Wireless Keyboard MK550')
_D(
    'Wireless Keyboard MK700',
    protocol=1.0,
    wpid='2008',
    registers=(_R.battery_status, ),
    settings=[
        _RS.fn_swap(),
    ],
)
_D(
    'Wireless Solar Keyboard K750',
    protocol=2.0,
    wpid='4002',
    settings=[_FS.fn_swap()],
)
_D(
    'Wireless Multi-Device Keyboard K780',
    protocol=4.5,
    wpid='405B',
    settings=[_FS.new_fn_swap()],
)
_D(
    'Wireless Illuminated Keyboard K800',
    protocol=1.0,
    wpid='2010',
    registers=(
        _R.battery_status,
        _R.three_leds,
    ),
    settings=[
        _RS.fn_swap(),
        _RS.hand_detection(),
    ],
)
_D(
    'Wireless Illuminated Keyboard K800 new',
    codename='K800 new',
    protocol=4.5,
    wpid='406E',
    settings=[_FS.fn_swap()],
)
_D(
    'Illuminated Living-Room Keyboard K830',
    protocol=2.0,
    wpid='4032',
    settings=[_FS.new_fn_swap()],
)
_D('Craft Advanced Keyboard', codename='Craft', protocol=4.5, wpid='4066')

_D(
    'Wireless Keyboard S510',
    codename='S510',
    protocol=1.0,
    wpid='3622',
    registers=(_R.battery_status, ),
)
_D(
    'Wireless Keyboard EX100',
    codename='EX100',
    protocol=1.0,
    wpid='6500',
    registers=(_R.battery_status, ),
)

# Mice

_D('Wireless Mouse M150', protocol=2.0, wpid='4022')
_D('Wireless Mouse M175', protocol=2.0, wpid='4008')
_D(
    'Wireless Mouse M185 new',
    codename='M185n',
    protocol=4.5,
    wpid='4054',
    settings=[
        _FS.lowres_smooth_scroll(),
        _FS.pointer_speed(),
    ]
)
# Apparently Logitech uses wpid 4055 for three different mice
# That's not so strange, as M185 is used on both Unifying-ready and non-Unifying-ready mice
_D(
    'Wireless Mouse M185/M235/M310',
    codename='M185/M235/M310',
    protocol=4.5,
    wpid='4055',
    settings=[
        _FS.lowres_smooth_scroll(),
        _FS.pointer_speed(),
    ]
)
_D('Wireless Mouse M185', protocol=2.0, wpid='4038')
_D('Wireless Mouse M187', protocol=2.0, wpid='4019')
_D('Wireless Mouse M215', protocol=1.0, wpid='1020')
_D(
    'Wireless Mouse M305',
    protocol=1.0,
    wpid='101F',
    registers=(_R.battery_status, ),
    settings=[
        _RS.side_scroll(),
    ],
)
_D(
    'Wireless Mouse M310',
    protocol=1.0,
    wpid='1024',
    registers=(_R.battery_status, ),
)
_D('Wireless Mouse M315')
_D('Wireless Mouse M317')
_D('Wireless Mouse M325', protocol=2.0, wpid='400A', settings=[
    _FS.hi_res_scroll(),
])
_D('Wireless Mouse M345', protocol=2.0, wpid='4017')
_D(
    'Wireless Mouse M350',
    protocol=1.0,
    wpid='101C',
    registers=(_R.battery_charge, ),
)
_D('Wireless Mouse Pebble M350', codename='Pebble', protocol=2.0, wpid='4080')
_D(
    'Wireless Mouse M505',
    codename='M505/B605',
    protocol=1.0,
    wpid='101D',
    registers=(_R.battery_charge, ),
    settings=[
        _RS.smooth_scroll(),
        _RS.side_scroll(),
    ],
)
_D(
    'Wireless Mouse M510',
    protocol=1.0,
    wpid='1025',
    registers=(_R.battery_status, ),
    settings=[
        # _RS.smooth_scroll(),	# writing the bit to the register doesn't cause an error, but the bit doesn't turn on
        _RS.side_scroll(),
    ],
)
_D('Wireless Mouse M510', codename='M510v2', protocol=2.0, wpid='4051', settings=[
    _FS.lowres_smooth_scroll(),
])
_D('Couch Mouse M515', protocol=2.0, wpid='4007')
_D('Wireless Mouse M525', protocol=2.0, wpid='4013')
_D(
    'Multi Device Silent Mouse M585/M590',
    codename='M585/M590',
    protocol=4.5,
    wpid='406B',
    settings=[
        _FS.lowres_smooth_scroll(),
        _FS.pointer_speed(),
    ],
)
_D('Touch Mouse M600', protocol=2.0, wpid='401A')
_D(
    'Marathon Mouse M705 (M-R0009)',
    codename='M705 (M-R0009)',
    protocol=1.0,
    wpid='101B',
    registers=(_R.battery_charge, ),
    settings=[
        _RS.smooth_scroll(),
        _RS.side_scroll(),
    ],
)
_D(
    'Marathon Mouse M705 (M-R0073)',
    codename='M705 (M-R0073)',
    protocol=4.5,
    wpid='406D',
    settings=[
        _FS.hires_smooth_invert(),
        _FS.hires_smooth_resolution(),
        _FS.pointer_speed(),
    ]
)
_D('Zone Touch Mouse T400')
_D('Touch Mouse T620', protocol=2.0)
_D('Logitech Cube', kind=_DK.mouse, protocol=2.0)
_D(
    'Anywhere Mouse MX',
    codename='Anywhere MX',
    protocol=1.0,
    wpid='1017',
    registers=(_R.battery_charge, ),
    settings=[
        _RS.smooth_scroll(),
        _RS.side_scroll(),
    ],
)
_D(
    'Anywhere Mouse MX 2',
    codename='Anywhere MX 2',
    protocol=4.5,
    wpid='404A',
    settings=[
        _FS.hires_smooth_invert(),
        _FS.hires_smooth_resolution(),
    ],
)
_D(
    'Performance Mouse MX',
    codename='Performance MX',
    protocol=1.0,
    wpid='101A',
    registers=(
        _R.battery_status,
        _R.three_leds,
    ),
    settings=[
        _RS.dpi(choices=_PERFORMANCE_MX_DPIS),
        _RS.smooth_scroll(),
        _RS.side_scroll(),
    ],
)

_D(
    'Wireless Mouse MX Master',
    codename='MX Master',
    protocol=4.5,
    wpid='4041',
    settings=[
        _FS.hires_smooth_invert(),
        _FS.hires_smooth_resolution(),
    ],
)

_D(
    'Wireless Mouse MX Master 2S',
    codename='MX Master 2S',
    protocol=4.5,
    wpid='4069',
    settings=[
        _FS.hires_smooth_invert(),
        _FS.hires_smooth_resolution(),
    ],
)

_D('Wireless Mouse MX Vertical', codename='MX Vertical', protocol=4.5, wpid='407B')

_D(
    'G7 Cordless Laser Mouse',
    codename='G7',
    protocol=1.0,
    wpid='1002',
    registers=(_R.battery_status, ),
)
_D(
    'G700 Gaming Mouse',
    codename='G700',
    protocol=1.0,
    wpid='1023',
    registers=(
        _R.battery_status,
        _R.three_leds,
    ),
    settings=[
        _RS.smooth_scroll(),
        _RS.side_scroll(),
    ],
)
_D(
    'G700s Gaming Mouse',
    codename='G700s',
    protocol=1.0,
    wpid='102A',
    registers=(
        _R.battery_status,
        _R.three_leds,
    ),
    settings=[
        _RS.smooth_scroll(),
        _RS.side_scroll(),
    ],
)
_D(
    'LX5 Cordless Mouse',
    codename='LX5',
    protocol=1.0,
    wpid='5612',
    registers=(_R.battery_status, ),
)
_D(
    'Wireless Mouse M30',
    codename='M30',
    protocol=1.0,
    wpid='6822',
    registers=(_R.battery_status, ),
)
_D(
    'Wireless Mouse EX100',
    codename='EX100m',
    protocol=1.0,
    wpid='3F00',
    registers=(_R.battery_status, ),
    # settings=[ _RS.smooth_scroll(), ], # command accepted, but no change in whell action
)

# Trackballs

_D('Wireless Trackball M570')

# Touchpads

_D('Wireless Rechargeable Touchpad T650', protocol=2.0, wpid='4101')
_D('Wireless Touchpad', codename='Wireless Touch', protocol=2.0, wpid='4011')

#
# Classic Nano peripherals (that don't support the Unifying protocol).
# A wpid is necessary to properly identify them.
#

_D(
    'VX Nano Cordless Laser Mouse',
    codename='VX Nano',
    protocol=1.0,
    wpid=('100B', '100F'),
    registers=(_R.battery_charge, ),
    settings=[
        _RS.smooth_scroll(),
        _RS.side_scroll(),
    ],
)
_D(
    'V450 Nano Cordless Laser Mouse',
    codename='V450 Nano',
    protocol=1.0,
    wpid='1011',
    registers=(_R.battery_charge, ),
)
_D(
    'V550 Nano Cordless Laser Mouse',
    codename='V550 Nano',
    protocol=1.0,
    wpid='1013',
    registers=(_R.battery_charge, ),
    settings=[
        _RS.smooth_scroll(),
        _RS.side_scroll(),
    ],
)

# Mini receiver mice

_D(
    'MX610 Laser Cordless Mouse',
    codename='MX610',
    protocol=1.0,
    wpid='1001',
    registers=(_R.battery_status, ),
)
_D(
    'MX620 Laser Cordless Mouse',
    codename='MX620',
    protocol=1.0,
    wpid=('100A', '1016'),
    registers=(_R.battery_charge, ),
)
_D(
    'MX610 Left-Handled Mouse',
    codename='MX610L',
    protocol=1.0,
    wpid='1004',
    registers=(_R.battery_status, ),
)
_D(
    'V400 Laser Cordless Mouse',
    codename='V400',
    protocol=1.0,
    wpid='1003',
    registers=(_R.battery_status, ),
)
_D(
    'V450 Laser Cordless Mouse',
    codename='V450',
    protocol=1.0,
    wpid='1005',
    registers=(_R.battery_status, ),
)
_D(
    'VX Revolution',
    codename='VX Revolution',
    kind=_DK.mouse,
    protocol=1.0,
    wpid=('1006', '100D'),
    registers=(_R.battery_charge, ),
)
_D(
    'MX Air',
    codename='MX Air',
    protocol=1.0,
    kind=_DK.mouse,
    wpid=('1007', '100E'),
    registers=(_R.battery_charge, ),
)
_D(
    'MX Revolution',
    codename='MX Revolution',
    protocol=1.0,
    kind=_DK.mouse,
    wpid=('1008', '100C'),
    registers=(_R.battery_charge, ),
)
_D(
    'MX 1100 Cordless Laser Mouse',
    codename='MX 1100',
    protocol=1.0,
    kind=_DK.mouse,
    wpid='1014',
    registers=(_R.battery_charge, ),
    settings=[
        _RS.smooth_scroll(),
        _RS.side_scroll(),
    ],
)

# Some exotics...

_D('Fujitsu Sonic Mouse', codename='Sonic', protocol=1.0, wpid='1029')



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

from logging import DEBUG as _DEBUG
from logging import getLogger

from gi.repository import GLib, Gtk
from logitech_receiver.status import KEYS as _K
from solaar.i18n import _

from . import icons as _icons

_log = getLogger(__name__)
del getLogger

#
#
#
_PAIRING_TIMEOUT = 30  # seconds
_STATUS_CHECK = 500  # milliseconds


def _create_page(assistant, kind, header=None, icon_name=None, text=None):
    p = Gtk.VBox(False, 8)
    assistant.append_page(p)
    assistant.set_page_type(p, kind)

    if header:
        item = Gtk.HBox(False, 16)
        p.pack_start(item, False, True, 0)

        label = Gtk.Label(header)
        label.set_alignment(0, 0)
        label.set_line_wrap(True)
        item.pack_start(label, True, True, 0)

        if icon_name:
            icon = Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.DIALOG)
            icon.set_alignment(1, 0)
            item.pack_start(icon, False, False, 0)

    if text:
        label = Gtk.Label(text)
        label.set_alignment(0, 0)
        label.set_line_wrap(True)
        p.pack_start(label, False, False, 0)

    p.show_all()
    return p


def _check_lock_state(assistant, receiver, count=2):
    if not assistant.is_drawable():
        if _log.isEnabledFor(_DEBUG):
            _log.debug('assistant %s destroyed, bailing out', assistant)
        return False

    if receiver.status.get(_K.ERROR):
        # receiver.status.new_device = _fake_device(receiver)
        _pairing_failed(assistant, receiver, receiver.status.pop(_K.ERROR))
        return False

    if receiver.status.new_device:
        device, receiver.status.new_device = receiver.status.new_device, None
        _pairing_succeeded(assistant, receiver, device)
        return False

    if not receiver.status.lock_open:
        if count > 0:
            # the actual device notification may arrive after the lock was paired,
            # so have a little patience
            GLib.timeout_add(_STATUS_CHECK, _check_lock_state, assistant, receiver, count - 1)
        else:
            _pairing_failed(assistant, receiver, 'failed to open pairing lock')
        return False

    return True


def _prepare(assistant, page, receiver):
    index = assistant.get_current_page()
    if _log.isEnabledFor(_DEBUG):
        _log.debug('prepare %s %d %s', assistant, index, page)

    if index == 0:
        if receiver.set_lock(False, timeout=_PAIRING_TIMEOUT):
            assert receiver.status.new_device is None
            assert receiver.status.get(_K.ERROR) is None
            spinner = page.get_children()[-1]
            spinner.start()
            GLib.timeout_add(_STATUS_CHECK, _check_lock_state, assistant, receiver)
            assistant.set_page_complete(page, True)
        else:
            GLib.idle_add(_pairing_failed, assistant, receiver, 'the pairing lock did not open')
    else:
        assistant.remove_page(0)


def _finish(assistant, receiver):
    if _log.isEnabledFor(_DEBUG):
        _log.debug('finish %s', assistant)
    assistant.destroy()
    receiver.status.new_device = None
    if receiver.status.lock_open:
        receiver.set_lock()
    else:
        receiver.status[_K.ERROR] = None


def _pairing_failed(assistant, receiver, error):
    if _log.isEnabledFor(_DEBUG):
        _log.debug('%s fail: %s', receiver, error)

    assistant.commit()

    header = _('Pairing failed') + ': ' + _(str(error)) + '.'
    if 'timeout' in str(error):
        text = _('Make sure your device is within range, and has a decent battery charge.')
    elif str(error) == 'device not supported':
        text = _('A new device was detected, but it is not compatible with this receiver.')
    elif 'many' in str(error):
        text = _('The receiver only supports %d paired device(s).')
    else:
        text = _('No further details are available about the error.')
    _create_page(assistant, Gtk.AssistantPageType.SUMMARY, header, 'dialog-error', text)

    assistant.next_page()
    assistant.commit()


def _pairing_succeeded(assistant, receiver, device):
    assert device
    if _log.isEnabledFor(_DEBUG):
        _log.debug('%s success: %s', receiver, device)

    page = _create_page(assistant, Gtk.AssistantPageType.SUMMARY)

    header = Gtk.Label(_('Found a new device:'))
    header.set_alignment(0.5, 0)
    page.pack_start(header, False, False, 0)

    device_icon = Gtk.Image()
    icon_set = _icons.device_icon_set(device.name, device.kind)
    device_icon.set_from_icon_set(icon_set, Gtk.IconSize.LARGE)
    device_icon.set_alignment(0.5, 1)
    page.pack_start(device_icon, True, True, 0)

    device_label = Gtk.Label()
    device_label.set_markup('<b>%s</b>' % device.name)
    device_label.set_alignment(0.5, 0)
    page.pack_start(device_label, True, True, 0)

    hbox = Gtk.HBox(False, 8)
    hbox.pack_start(Gtk.Label(' '), False, False, 0)
    hbox.set_property('expand', False)
    hbox.set_property('halign', Gtk.Align.CENTER)
    page.pack_start(hbox, False, False, 0)

    def _check_encrypted(dev):
        if assistant.is_drawable():
            if device.status.get(_K.LINK_ENCRYPTED) is False:
                hbox.pack_start(Gtk.Image.new_from_icon_name('security-low', Gtk.IconSize.MENU), False, False, 0)
                hbox.pack_start(Gtk.Label(_('The wireless link is not encrypted') + '!'), False, False, 0)
                hbox.show_all()
            else:
                return True

    GLib.timeout_add(_STATUS_CHECK, _check_encrypted, device)

    page.show_all()

    assistant.next_page()
    assistant.commit()


def create(receiver):
    assert receiver is not None
    assert receiver.kind is None

    assistant = Gtk.Assistant()
    assistant.set_title(_('%(receiver_name)s: pair new device') % {'receiver_name': receiver.name})
    assistant.set_icon_name('list-add')

    assistant.set_size_request(400, 240)
    assistant.set_resizable(False)
    assistant.set_role('pair-device')

    page_text = _('If the device is already turned on, turn if off and on again.')
    if receiver.remaining_pairings() and receiver.remaining_pairings() >= 0:
        page_text += _('\n\nThis receiver has %d pairing(s) remaining.') % receiver.remaining_pairings()
        page_text += _('\nCancelling at this point will not use up a pairing.')

    page_intro = _create_page(
        assistant, Gtk.AssistantPageType.PROGRESS, _('Turn on the device you want to pair.'),
        'preferences-desktop-peripherals', page_text
    )
    spinner = Gtk.Spinner()
    spinner.set_visible(True)
    page_intro.pack_end(spinner, True, True, 24)

    assistant.connect('prepare', _prepare, receiver)
    assistant.connect('cancel', _finish, receiver)
    assistant.connect('close', _finish, receiver)

    return assistant



#!/usr/bin/env python3

from glob import glob as _glob

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

# from solaar import NAME, __version__
__version__ = '1.0.3rc2'
NAME = 'Solaar'


def _data_files():
    from os.path import dirname as _dirname

    yield 'share/solaar/icons', _glob('share/solaar/icons/solaar*.svg')
    yield 'share/solaar/icons', _glob('share/solaar/icons/light_*.png')
    yield 'share/icons/hicolor/scalable/apps', ['share/solaar/icons/solaar.svg']

    for mo in _glob('share/locale/*/LC_MESSAGES/solaar.mo'):
        yield _dirname(mo), [mo]

    yield 'share/applications', ['share/applications/solaar.desktop']
    yield 'share/solaar/udev-rules.d', ['rules.d/42-logitech-unify-permissions.rules']

    del _dirname


setup(
    name=NAME.lower(),
    version=__version__,
    description='Linux devices manager for the Logitech Unifying Receiver.',
    long_description='''
Solaar is a Linux device manager for Logitech's Unifying Receiver peripherals.
It is able to pair/unpair devices with the receiver, for many devices show
battery status, and show and modify some of the modifiable features of devices.
'''.strip(),
    author='Daniel Pavel',
    license='GPLv2',
    url='http://pwr-solaar.github.io/Solaar/',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: X11 Applications :: GTK',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'License :: DFSG approved',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3 :: Only',
        'Operating System :: POSIX :: Linux',
        'Topic :: Utilities',
    ],
    platforms=['linux'],

    # sudo apt install python-gi python3-gi \
    #        gir1.2-gtk-3.0 gir1.2-notify-0.7 gir1.2-ayatanaappindicator3-0.1
    # os_requires=['gi.repository.GObject (>= 2.0)', 'gi.repository.Gtk (>= 3.0)'],
    python_requires='>=3.6',
    install_requires=[
        'pyudev (>= 0.13)',
    ],
    package_dir={'': 'lib'},
    packages=['hidapi', 'logitech_receiver', 'solaar', 'solaar.ui', 'solaar.cli'],
    data_files=list(_data_files()),
    scripts=_glob('bin/*'),
)



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Reprogrammable keys information
# Mostly from Logitech documentation, but with some edits for better Lunix compatability

from __future__ import absolute_import, division, print_function, unicode_literals

from .common import NamedInts as _NamedInts

# <controls.xml awk -F\" '/<Control /{sub(/^LD_FINFO_(CTRLID_)?/, "", $2);printf("\t%s=0x%04X,\n", $2, $4)}' | sort -t= -k2
CONTROL = _NamedInts(
    Volume_Up=0x0001,
    Volume_Down=0x0002,
    Mute=0x0003,
    Play__Pause=0x0004,
    Next=0x0005,
    Previous=0x0006,
    Stop=0x0007,
    Application_Switcher=0x0008,
    BURN=0x0009,
    Calculator=0x000A,
    CALENDAR=0x000B,
    CLOSE=0x000C,
    EJECT=0x000D,
    Mail=0x000E,
    HELP_AS_HID=0x000F,
    HELP_AS_F1=0x0010,
    LAUNCH_WORD_PROC=0x0011,
    LAUNCH_SPREADSHEET=0x0012,
    LAUNCH_PRESENTATION=0x0013,
    UNDO_AS_CTRL_Z=0x0014,
    UNDO_AS_HID=0x0015,
    REDO_AS_CTRL_Y=0x0016,
    REDO_AS_HID=0x0017,
    PRINT_AS_CTRL_P=0x0018,
    PRINT_AS_HID=0x0019,
    SAVE_AS_CTRL_S=0x001A,
    SAVE_AS_HID=0x001B,
    PRESET_A=0x001C,
    PRESET_B=0x001D,
    PRESET_C=0x001E,
    PRESET_D=0x001F,
    FAVORITES=0x0020,
    GADGETS=0x0021,
    MY_HOME=0x0022,
    GADGETS_AS_WIN_G=0x0023,
    MAXIMIZE_AS_HID=0x0024,
    MAXIMIZE_AS_WIN_SHIFT_M=0x0025,
    MINIMIZE_AS_HID=0x0026,
    MINIMIZE_AS_WIN_M=0x0027,
    MEDIA_PLAYER=0x0028,
    MEDIA_CENTER_LOGI=0x0029,
    MEDIA_CENTER_MSFT=0x002A,  # Should not be used as it is not reprogrammable under Windows
    CUSTOM_MENU=0x002B,
    MESSENGER=0x002C,
    MY_DOCUMENTS=0x002D,
    MY_MUSIC=0x002E,
    WEBCAM=0x002F,
    MY_PICTURES=0x0030,
    MY_VIDEOS=0x0031,
    MY_COMPUTER_AS_HID=0x0032,
    MY_COMPUTER_AS_WIN_E=0x0033,
    LAUNC_PICTURE_VIEWER=0x0035,
    ONE_TOUCH_SEARCH=0x0036,
    PRESET_1=0x0037,
    PRESET_2=0x0038,
    PRESET_3=0x0039,
    PRESET_4=0x003A,
    RECORD=0x003B,
    INTERNET_REFRESH=0x003C,
    ROTATE_RIGHT=0x003D,
    Search=0x003E,  # SEARCH
    SHUFFLE=0x003F,
    SLEEP=0x0040,
    INTERNET_STOP=0x0041,
    SYNCHRONIZE=0x0042,
    ZOOM=0x0043,
    ZOOM_IN_AS_HID=0x0044,
    ZOOM_IN_AS_CTRL_WHEEL=0x0045,
    ZOOM_IN_AS_CLTR_PLUS=0x0046,
    ZOOM_OUT_AS_HID=0x0047,
    ZOOM_OUT_AS_CTRL_WHEEL=0x0048,
    ZOOM_OUT_AS_CLTR_MINUS=0x0049,
    ZOOM_RESET=0x004A,
    ZOOM_FULL_SCREEN=0x004B,
    PRINT_SCREEN=0x004C,
    PAUSE_BREAK=0x004D,
    SCROLL_LOCK=0x004E,
    CONTEXTUAL_MENU=0x004F,
    Left_Button=0x0050,  # LEFT_CLICK
    Right_Button=0x0051,  # RIGHT_CLICK
    Middle_Button=0x0052,  # MIDDLE_BUTTON
    Back_Button=0x0053,  # from M510v2 was BACK_AS_BUTTON_4
    Back=0x0054,  # BACK_AS_HID
    BACK_AS_ALT_WIN_ARROW=0x0055,
    Forward_Button=0x0056,  # from M510v2 was FORWARD_AS_BUTTON_5
    FORWARD_AS_HID=0x0057,
    FORWARD_AS_ALT_WIN_ARROW=0x0058,
    BUTTON_6=0x0059,
    LEFT_SCROLL_AS_BUTTON_7=0x005A,
    Left_Tilt=0x005B,  # from M510v2 was LEFT_SCROLL_AS_AC_PAN
    RIGHT_SCROLL_AS_BUTTON_8=0x005C,
    Right_Tilt=0x005D,  # from M510v2 was RIGHT_SCROLL_AS_AC_PAN
    BUTTON_9=0x005E,
    BUTTON_10=0x005F,
    BUTTON_11=0x0060,
    BUTTON_12=0x0061,
    BUTTON_13=0x0062,
    BUTTON_14=0x0063,
    BUTTON_15=0x0064,
    BUTTON_16=0x0065,
    BUTTON_17=0x0066,
    BUTTON_18=0x0067,
    BUTTON_19=0x0068,
    BUTTON_20=0x0069,
    BUTTON_21=0x006A,
    BUTTON_22=0x006B,
    BUTTON_23=0x006C,
    BUTTON_24=0x006D,
    Show_Desktop=0x006E,  # Show_Desktop
    Lock_PC=0x006F,
    FN_F1=0x0070,
    FN_F2=0x0071,
    FN_F3=0x0072,
    FN_F4=0x0073,
    FN_F5=0x0074,
    FN_F6=0x0075,
    FN_F7=0x0076,
    FN_F8=0x0077,
    FN_F9=0x0078,
    FN_F10=0x0079,
    FN_F11=0x007A,
    FN_F12=0x007B,
    FN_F13=0x007C,
    FN_F14=0x007D,
    FN_F15=0x007E,
    FN_F16=0x007F,
    FN_F17=0x0080,
    FN_F18=0x0081,
    FN_F19=0x0082,
    IOS_HOME=0x0083,
    ANDROID_HOME=0x0084,
    ANDROID_MENU=0x0085,
    ANDROID_SEARCH=0x0086,
    ANDROID_BACK=0x0087,
    HOME_COMBO=0x0088,
    LOCK_COMBO=0x0089,
    IOS_VIRTUAL_KEYBOARD=0x008A,
    IOS_LANGUAGE_SWICH=0x008B,
    MAC_EXPOSE=0x008C,
    MAC_DASHBOARD=0x008D,
    WIN7_SNAP_LEFT=0x008E,
    WIN7_SNAP_RIGHT=0x008F,
    Minimize_Window=0x0090,  # WIN7_MINIMIZE_AS_WIN_ARROW
    Maximize_Window=0x0091,  # WIN7_MAXIMIZE_AS_WIN_ARROW
    WIN7_STRETCH_UP=0x0092,
    WIN7_MONITOR_SWITCH_AS_WIN_SHIFT_LEFTARROW=0x0093,
    WIN7_MONITOR_SWITCH_AS_WIN_SHIFT_RIGHTARROW=0x0094,
    Switch_Screen=0x0095,  # WIN7_SHOW_PRESENTATION_MODE
    WIN7_SHOW_MOBILITY_CENTER=0x0096,
    ANALOG_HSCROLL=0x0097,
    METRO_APPSWITCH=0x009F,
    METRO_APPBAR=0x00A0,
    METRO_CHARMS=0x00A1,
    CALC_VKEYBOARD=0x00A2,
    METRO_SEARCH=0x00A3,
    COMBO_SLEEP=0x00A4,
    METRO_SHARE=0x00A5,
    METRO_SETTINGS=0x00A6,
    METRO_DEVICES=0x00A7,
    METRO_START_SCREEN=0x00A9,
    ZOOMIN=0x00AA,
    ZOOMOUT=0x00AB,
    BACK_HSCROLL=0x00AC,
    SHOW_DESKTOP_HPP=0x00AE,
    Fn_Left_Click=0x00B7,  # from K400 Plus
    # https://docs.google.com/document/u/0/d/1YvXICgSe8BcBAuMr4Xu_TutvAxaa-RnGfyPFWBWzhkc/export?format=docx
    # Extract to csv.  Eliminate extra linefeeds and spaces.
    # awk -F, '/0x/{gsub(" \\+ ","_",$2); gsub("/","__",$2); gsub(" -","_Down",$2);
    # gsub(" \\+","_Up",$2); gsub("[()\"-]","",$2); gsub(" ","_",$2); printf("\t%s=0x%04X,\n", $2, $1)}' < controls.cvs
    Second_Left_Click=0x00B8,  # Second_LClick / on K400 Plus
    Fn_Second_Left_Click=0x00B9,  # Fn_Second_LClick
    MultiPlatform_App_Switch=0x00BA,
    MultiPlatform_Home=0x00BB,
    MultiPlatform_Menu=0x00BC,
    MultiPlatform_Back=0x00BD,
    MultiPlatform_Insert=0x00BE,
    Screen_Capture__Print_Screen=0x00BF,  # on Craft Keyboard
    Fn_Down=0x00C0,
    Fn_Up=0x00C1,
    Multiplatform_Lock=0x00C2,
    App_Switch_Gesture=0x00C3,  # Thumb_Button on MX Master
    Smart_Shift=0x00C4,  # Top_Button on MX Master
    Microphone=0x00C5,
    Wifi=0x00C6,
    Brightness_Down=0x00C7,
    Brightness_Up=0x00C8,
    Display_out__project_screen_=0x00C9,
    View_Open_Apps=0x00CA,
    View_all_apps=0x00CB,
    Switch_App=0x00CC,
    Fn_inversion_change=0x00CD,
    MultiPlatform_back=0x00CE,
    Multiplatform_forward=0x00CF,
    Multiplatform_gesture_button=0x00D0,
    Host_Switch_channel_1=0x00D1,
    Host_Switch_channel_2=0x00D2,
    Host_Switch_channel_3=0x00D3,
    Multiplatform_search=0x00D4,
    Multiplatform_Home__Mission_Control=0x00D5,
    Multiplatform_Menu__Show__Hide_Virtual_Keyboard__Launchpad=0x00D6,
    Virtual_Gesture_Button=0x00D7,
    Cursor_Button_Long_press=0x00D8,
    Next_Button_Shortpress=0x00D9,  # Next_Button
    Next_Button_Longpress=0x00DA,
    Back_Button_Shortpress=0x00DB,  # Back
    Back_Button_Longpress=0x00DC,
    Multi_Platform_Language_Switch=0x00DD,
    F_Lock=0x00DE,
    Switch_Highlight=0x00DF,
    Mission_Control__Task_View=0x00E0,  # Switch_Workspaces on Craft Keyboard
    Dashboard_Launchpad__Action_Center=0x00E1,  # Application_Launcher on Craft Keyboard
    Backlight_Down=0x00E2,
    Backlight_Up=0x00E3,
    Previous_Fn=0x00E4,  # Reprogrammable_Previous_Track / on Craft Keyboard
    Play__Pause_Fn=0x00E5,  # Reprogrammable_Play__Pause / on Craft Keyboard
    Next_Fn=0x00E6,  # Reprogrammable_Next_Track / on Craft Keyboard
    Mute_Fn=0x00E7,  # Reprogrammable_Mute / on Craft Keyboard
    Volume_Down_Fn=0x00E8,  # Reprogrammable_Volume_Down / on Craft Keyboard
    Volume_Up_Fn=0x00E9,  # Reprogrammable_Volume_Up / on Craft Keyboard
    App_Contextual_Menu__Right_Click=0x00EA,  # Context_Menu on Craft Keyboard
    Right_Arrow=0x00EB,
    Left_Arrow=0x00EC,
    DPI_Change=0x00ED,
    New_Tab=0x00EE,
    F2=0x00EF,
    F3=0x00F0,
    F4=0x00F1,
    F5=0x00F2,
    F6=0x00F3,
    F7=0x00F4,
    F8=0x00F5,
    F1=0x00F6,
    Next_Color_Effect=0x00F7,
    Increase_Color_Effect_Speed=0x00F8,
    Decrease_Color_Effect_Speed=0x00F9,
    Load_Lighting_Custom_Profile=0x00FA,
    Laser_button_short_press=0x00FB,
    Laser_button_long_press=0x00FC,
    DPI_switch=0x00FD,
    MultiPlatform_Home__Show_Desktop=0x00FE,
    MultiPlatform_App_Switch__Show_Dashboard=0x00FF,
    MultiPlatform_App_Switch_2=0x0100,  # MultiPlatform_App_Switch
    Fn_Inversion__Hot_Key=0x0101,
    LeftAndRightClick=0x0102,
    LED_TOGGLE=0x013B,  #
)
CONTROL._fallback = lambda x: 'unknown:%04X' % x

# <tasks.xml awk -F\" '/<Task /{gsub(/ /, "_", $6); printf("\t%s=0x%04X,\n", $6, $4)}'
TASK = _NamedInts(
    Volume_Up=0x0001,
    Volume_Down=0x0002,
    Mute=0x0003,
    # Multimedia tasks:
    Play__Pause=0x0004,
    Next=0x0005,
    Previous=0x0006,
    Stop=0x0007,
    Application_Switcher=0x0008,
    BurnMediaPlayer=0x0009,
    Calculator=0x000A,
    Calendar=0x000B,
    Close_Application=0x000C,
    Eject=0x000D,
    Email=0x000E,
    Help=0x000F,
    OffDocument=0x0010,
    OffSpreadsheet=0x0011,
    OffPowerpnt=0x0012,
    Undo=0x0013,
    Redo=0x0014,
    Print=0x0015,
    Save=0x0016,
    SmartKeySet=0x0017,
    Favorites=0x0018,
    GadgetsSet=0x0019,
    HomePage=0x001A,
    WindowsRestore=0x001B,
    WindowsMinimize=0x001C,
    Music=0x001D,  # also known as MediaPlayer

    # Both 0x001E and 0x001F are known as MediaCenterSet
    Media_Center_Logitech=0x001E,
    Media_Center_Microsoft=0x001F,
    UserMenu=0x0020,
    Messenger=0x0021,
    PersonalFolders=0x0022,
    MyMusic=0x0023,
    Webcam=0x0024,
    PicturesFolder=0x0025,
    MyVideos=0x0026,
    My_Computer=0x0027,
    PictureAppSet=0x0028,
    Search=0x0029,  # also known as AdvSmartSearch
    RecordMediaPlayer=0x002A,
    BrowserRefresh=0x002B,
    RotateRight=0x002C,
    Search_Files=0x002D,  # SearchForFiles
    MM_SHUFFLE=0x002E,
    Sleep=0x002F,  # also known as StandBySet
    BrowserStop=0x0030,
    OneTouchSync=0x0031,
    ZoomSet=0x0032,
    ZoomBtnInSet2=0x0033,
    ZoomBtnInSet=0x0034,
    ZoomBtnOutSet2=0x0035,
    ZoomBtnOutSet=0x0036,
    ZoomBtnResetSet=0x0037,
    Left_Click=0x0038,  # LeftClick
    Right_Click=0x0039,  # RightClick
    Mouse_Middle_Button=0x003A,  # from M510v2 was MiddleMouseButton
    Back=0x003B,
    Mouse_Back_Button=0x003C,  # from M510v2 was BackEx
    BrowserForward=0x003D,
    Mouse_Forward_Button=0x003E,  # from M510v2 was BrowserForwardEx
    Mouse_Scroll_Left_Button_=0x003F,  # from M510v2 was HorzScrollLeftSet
    Mouse_Scroll_Right_Button=0x0040,  # from M510v2 was HorzScrollRightSet
    QuickSwitch=0x0041,
    BatteryStatus=0x0042,
    Show_Desktop=0x0043,  # ShowDesktop
    WindowsLock=0x0044,
    FileLauncher=0x0045,
    FolderLauncher=0x0046,
    GotoWebAddress=0x0047,
    GenericMouseButton=0x0048,
    KeystrokeAssignment=0x0049,
    LaunchProgram=0x004A,
    MinMaxWindow=0x004B,
    VOLUMEMUTE_NoOSD=0x004C,
    New=0x004D,
    Copy=0x004E,
    CruiseDown=0x004F,
    CruiseUp=0x0050,
    Cut=0x0051,
    Do_Nothing=0x0052,
    PageDown=0x0053,
    PageUp=0x0054,
    Paste=0x0055,
    SearchPicture=0x0056,
    Reply=0x0057,
    PhotoGallerySet=0x0058,
    MM_REWIND=0x0059,
    MM_FASTFORWARD=0x005A,
    Send=0x005B,
    ControlPanel=0x005C,
    UniversalScroll=0x005D,
    AutoScroll=0x005E,
    GenericButton=0x005F,
    MM_NEXT=0x0060,
    MM_PREVIOUS=0x0061,
    Do_Nothing_One=0x0062,  # also known as Do_Nothing
    SnapLeft=0x0063,
    SnapRight=0x0064,
    WinMinRestore=0x0065,
    WinMaxRestore=0x0066,
    WinStretch=0x0067,
    SwitchMonitorLeft=0x0068,
    SwitchMonitorRight=0x0069,
    ShowPresentation=0x006A,
    ShowMobilityCenter=0x006B,
    HorzScrollNoRepeatSet=0x006C,
    TouchBackForwardHorzScroll=0x0077,
    MetroAppSwitch=0x0078,
    MetroAppBar=0x0079,
    MetroCharms=0x007A,
    Calculator_VKEY=0x007B,  # also known as Calculator
    MetroSearch=0x007C,
    MetroStartScreen=0x0080,
    MetroShare=0x007D,
    MetroSettings=0x007E,
    MetroDevices=0x007F,
    MetroBackLeftHorz=0x0082,
    MetroForwRightHorz=0x0083,
    Win8_Back=0x0084,  # also known as MetroCharms
    Win8_Forward=0x0085,  # also known as AppSwitchBar
    Win8Charm_Appswitch_GifAnimation=0x0086,
    Win8BackHorzLeft=0x008B,  # also known as Back
    Win8ForwardHorzRight=0x008C,  # also known as BrowserForward
    MetroSearch2=0x0087,
    MetroShare2=0x0088,
    MetroSettings2=0x008A,
    MetroDevices2=0x0089,
    Win8MetroWin7Forward=0x008D,  # also known as MetroStartScreen
    Win8ShowDesktopWin7Back=0x008E,  # also known as ShowDesktop
    MetroApplicationSwitch=0x0090,  # also known as MetroStartScreen
    ShowUI=0x0092,
    # https://docs.google.com/document/d/1Dpx_nWRQAZox_zpZ8SNc9nOkSDE9svjkghOCbzopabc/edit
    # Extract to csv.  Eliminate extra linefeeds and spaces. Turn / into __ and space into _
    # awk -F, '/0x/{gsub(" \\+ ","_",$2);  gsub("_-","_Down",$2); gsub("_\\+","_Up",$2);
    # gsub("[()\"-]","",$2); gsub(" ","_",$2); printf("\t%s=0x%04X,\n", $2, $1)}' < tasks.csv > tasks.py
    Switch_Presentation__Switch_Screen=0x0093,  # on K400 Plus
    Minimize_Window=0x0094,
    Maximize_Window=0x0095,  # on K400 Plus
    MultiPlatform_App_Switch=0x0096,
    MultiPlatform_Home=0x0097,
    MultiPlatform_Menu=0x0098,
    MultiPlatform_Back=0x0099,
    Switch_Language=0x009A,  # Mac_switch_language
    Screen_Capture=0x009B,  # Mac_screen_Capture, on Craft Keyboard
    Gesture_Button=0x009C,
    Smart_Shift=0x009D,
    AppExpose=0x009E,
    Smart_Zoom=0x009F,
    Lookup=0x00A0,
    Microphone_on__off=0x00A1,
    Wifi_on__off=0x00A2,
    Brightness_Down=0x00A3,
    Brightness_Up=0x00A4,
    Display_Out=0x00A5,
    View_Open_Apps=0x00A6,
    View_All_Open_Apps=0x00A7,
    AppSwitch=0x00A8,
    Gesture_Button_Navigation=0x00A9,  # Mouse_Thumb_Button on MX Master
    Fn_inversion=0x00AA,
    Multiplatform_Back=0x00AB,
    Multiplatform_Forward=0x00AC,
    Multiplatform_Gesture_Button=0x00AD,
    HostSwitch_Channel_1=0x00AE,
    HostSwitch_Channel_2=0x00AF,
    HostSwitch_Channel_3=0x00B0,
    Multiplatform_Search=0x00B1,
    Multiplatform_Home__Mission_Control=0x00B2,
    Multiplatform_Menu__Launchpad=0x00B3,
    Virtual_Gesture_Button=0x00B4,
    Cursor=0x00B5,
    Keyboard_Right_Arrow=0x00B6,
    SW_Custom_Highlight=0x00B7,
    Keyboard_Left_Arrow=0x00B8,
    TBD=0x00B9,
    Multiplatform_Language_Switch=0x00BA,
    SW_Custom_Highlight_2=0x00BB,
    Fast_Forward=0x00BC,
    Fast_Backward=0x00BD,
    Switch_Highlighting=0x00BE,
    Mission_Control__Task_View=0x00BF,  # Switch_Workspace on Craft Keyboard
    Dashboard_Launchpad__Action_Center=0x00C0,  # Application_Launcher on Craft Keyboard
    Backlight_Down=0x00C1,  # Backlight_Down_FW_internal_function
    Backlight_Up=0x00C2,  # Backlight_Up_FW_internal_function
    Right_Click__App_Contextual_Menu=0x00C3,  # Context_Menu on Craft Keyboard
    DPI_Change=0x00C4,
    New_Tab=0x00C5,
    F2=0x00C6,
    F3=0x00C7,
    F4=0x00C8,
    F5=0x00C9,
    F6=0x00CA,
    F7=0x00CB,
    F8=0x00CC,
    F1=0x00CD,
    Laser_Button=0x00CE,
    Laser_Button_Long_Press=0x00CF,
    Start_Presentation=0x00D0,
    Blank_Screen=0x00D1,
    DPI_Switch=0x00D2,  # AdjustDPI on MX Vertical
    Home__Show_Desktop=0x00D3,
    App_Switch__Dashboard=0x00D4,
    App_Switch=0x00D5,
    Fn_Inversion=0x00D6,
    LeftAndRightClick=0x00D7,
    LedToggle=0x00DD,  #
)
TASK._fallback = lambda x: 'unknown:%04X' % x
# hidpp 4.5 info from https://lekensteyn.nl/files/logitech/x1b04_specialkeysmsebuttons.html
KEY_FLAG = _NamedInts(
    virtual=0x80,
    persistently_divertable=0x40,
    divertable=0x20,
    reprogrammable=0x10,
    FN_sensitive=0x08,
    nonstandard=0x04,
    is_FN=0x02,
    mse=0x01
)

DISABLE = _NamedInts(
    Caps_Lock=0x01,
    Num_Lock=0x02,
    Scroll_Lock=0x04,
    Insert=0x08,
    Win=0x10,  # aka Super
)
DISABLE._fallback = lambda x: 'unknown:%02X' % x



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

import threading as _threading

from logging import DEBUG as _DEBUG
from logging import INFO as _INFO
from logging import getLogger

from . import base as _base

# from time import time as _timestamp

# for both Python 2 and 3
try:
    from Queue import Queue as _Queue
except ImportError:
    from queue import Queue as _Queue

_log = getLogger(__name__)
del getLogger

#
#
#


class _ThreadedHandle(object):
    """A thread-local wrapper with different open handles for each thread.

    Closing a ThreadedHandle will close all handles.
    """

    __slots__ = ('path', '_local', '_handles', '_listener')

    def __init__(self, listener, path, handle):
        assert listener is not None
        assert path is not None
        assert handle is not None
        assert isinstance(handle, int)

        self._listener = listener
        self.path = path
        self._local = _threading.local()
        # take over the current handle for the thread doing the replacement
        self._local.handle = handle
        self._handles = [handle]

    def _open(self):
        handle = _base.open_path(self.path)
        if handle is None:
            _log.error('%r failed to open new handle', self)
        else:
            # if _log.isEnabledFor(_DEBUG):
            #     _log.debug("%r opened new handle %d", self, handle)
            self._local.handle = handle
            self._handles.append(handle)
            return handle

    def close(self):
        if self._local:
            self._local = None
            handles, self._handles = self._handles, []
            if _log.isEnabledFor(_DEBUG):
                _log.debug('%r closing %s', self, handles)
            for h in handles:
                _base.close(h)

    @property
    def notifications_hook(self):
        if self._listener:
            assert isinstance(self._listener, _threading.Thread)
            if _threading.current_thread() == self._listener:
                return self._listener._notifications_hook

    def __del__(self):
        self._listener = None
        self.close()

    def __index__(self):
        if self._local:
            try:
                return self._local.handle
            except Exception:
                return self._open()

    __int__ = __index__

    def __str__(self):
        if self._local:
            return str(int(self))

    __unicode__ = __str__

    def __repr__(self):
        return '<_ThreadedHandle(%s)>' % self.path

    def __bool__(self):
        return bool(self._local)

    __nonzero__ = __bool__


#
#
#

# How long to wait during a read for the next packet, in seconds
# Ideally this should be rather long (10s ?), but the read is blocking
# and this means that when the thread is signalled to stop, it would take
# a while for it to acknowledge it.
# Forcibly closing the file handle on another thread does _not_ interrupt the
# read on Linux systems.
_EVENT_READ_TIMEOUT = 0.4  # in seconds

# After this many reads that did not produce a packet, call the tick() method.
# This only happens if tick_period is enabled (>0) for the Listener instance.
# _IDLE_READS = 1 + int(5 // _EVENT_READ_TIMEOUT)  # wait at least 5 seconds between ticks


class EventsListener(_threading.Thread):
    """Listener thread for notifications from the Unifying Receiver.

    Incoming packets will be passed to the callback function in sequence.
    """
    def __init__(self, receiver, notifications_callback):
        super(EventsListener, self).__init__(name=self.__class__.__name__ + ':' + receiver.path.split('/')[2])

        self.daemon = True
        self._active = False

        self.receiver = receiver
        self._queued_notifications = _Queue(16)
        self._notifications_callback = notifications_callback

        # self.tick_period = 0

    def run(self):
        self._active = True

        # replace the handle with a threaded one
        self.receiver.handle = _ThreadedHandle(self, self.receiver.path, self.receiver.handle)
        # get the right low-level handle for this thread
        ihandle = int(self.receiver.handle)
        if _log.isEnabledFor(_INFO):
            _log.info('started with %s (%d)', self.receiver, ihandle)

        self.has_started()

        # last_tick = 0
        # the first idle read -- delay it a bit, and make sure to stagger
        # idle reads for multiple receivers
        # idle_reads = _IDLE_READS + (ihandle % 5) * 2

        while self._active:
            if self._queued_notifications.empty():
                try:
                    # _log.debug("read next notification")
                    n = _base.read(ihandle, _EVENT_READ_TIMEOUT)
                except _base.NoReceiver:
                    _log.warning('receiver disconnected')
                    self.receiver.close()
                    break

                if n:
                    n = _base.make_notification(*n)
            else:
                # deliver any queued notifications
                n = self._queued_notifications.get()

            if n:
                # if _log.isEnabledFor(_DEBUG):
                #     _log.debug("%s: processing %s", self.receiver, n)
                try:
                    self._notifications_callback(n)
                except Exception:
                    _log.exception('processing %s', n)

            # elif self.tick_period:
            #     idle_reads -= 1
            #     if idle_reads <= 0:
            #         idle_reads = _IDLE_READS
            #         now = _timestamp()
            #         if now - last_tick >= self.tick_period:
            #             last_tick = now
            #             self.tick(now)

        del self._queued_notifications
        self.has_stopped()

    def stop(self):
        """Tells the listener to stop as soon as possible."""
        self._active = False

    def has_started(self):
        """Called right after the thread has started, and before it starts
        reading notification packets."""
        pass

    def has_stopped(self):
        """Called right before the thread stops."""
        pass

    # def tick(self, timestamp):
    #     """Called about every tick_period seconds."""
    #     pass

    def _notifications_hook(self, n):
        # Only consider unhandled notifications that were sent from this thread,
        # i.e. triggered by a callback handling a previous notification.
        assert _threading.current_thread() == self
        if self._active:  # and _threading.current_thread() == self:
            # if _log.isEnabledFor(_DEBUG):
            #     _log.debug("queueing unhandled %s", n)
            if not self._queued_notifications.full():
                self._queued_notifications.put(n)

    def __bool__(self):
        return bool(self._active and self.receiver)

    __nonzero__ = __bool__



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Some common functions and types.

from __future__ import absolute_import, division, print_function, unicode_literals

from binascii import hexlify as _hexlify
from collections import namedtuple
from struct import pack, unpack

try:
    unicode  # noqa: F821
    # if Python2, unicode_literals will mess our first (un)pack() argument
    _pack_str = pack
    _unpack_str = unpack
    pack = lambda x, *args: _pack_str(str(x), *args)
    unpack = lambda x, *args: _unpack_str(str(x), *args)

    is_string = lambda d: isinstance(d, unicode) or isinstance(d, str)  # noqa: F821
    # no easy way to distinguish between b'' and '' :(
    # or (isinstance(d, str) \
    #     and not any((chr(k) in d for k in range(0x00, 0x1F))) \
    #     and not any((chr(k) in d for k in range(0x80, 0xFF))) \
    #     )
except Exception:
    # this is certanly Python 3
    # In Py3, unicode and str are equal (the unicode object does not exist)
    is_string = lambda d: isinstance(d, str)

#
#
#


class NamedInt(int):
    """An reqular Python integer with an attached name.

    Caution: comparison with strings will also match this NamedInt's name
    (case-insensitive)."""
    def __new__(cls, value, name):
        assert is_string(name)
        obj = int.__new__(cls, value)
        obj.name = str(name)
        return obj

    def bytes(self, count=2):
        return int2bytes(self, count)

    def __eq__(self, other):
        if isinstance(other, NamedInt):
            return int(self) == int(other) and self.name == other.name
        if isinstance(other, int):
            return int(self) == int(other)
        if is_string(other):
            return self.name.lower() == other.lower()
        # this should catch comparisons with bytes in Py3
        if other is not None:
            raise TypeError('Unsupported type ' + str(type(other)))

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return int(self)

    def __str__(self):
        return self.name

    __unicode__ = __str__

    def __repr__(self):
        return 'NamedInt(%d, %r)' % (int(self), self.name)


class NamedInts(object):
    """An ordered set of NamedInt values.

    Indexing can be made by int or string, and will return the corresponding
    NamedInt if it exists in this set, or `None`.

    Extracting slices will return all present NamedInts in the given interval
    (extended slices are not supported).

    Assigning a string to an indexed int will create a new NamedInt in this set;
    if the value already exists in the set (int or string), ValueError will be
    raised.
    """
    __slots__ = ('__dict__', '_values', '_indexed', '_fallback')

    def __init__(self, **kwargs):
        def _readable_name(n):
            if not is_string(n):
                raise TypeError('expected (unicode) string, got ' + str(type(n)))
            return n.replace('__', '/').replace('_', ' ')

        # print (repr(kwargs))
        values = {k: NamedInt(v, _readable_name(k)) for (k, v) in kwargs.items()}
        self.__dict__ = values
        self._values = sorted(list(values.values()))
        self._indexed = {int(v): v for v in self._values}
        # assert len(values) == len(self._indexed)
        # "(%d) %r\n=> (%d) %r" % (len(values), values, len(self._indexed), self._indexed)
        self._fallback = None

    @classmethod
    def list(cls, items, name_generator=lambda x: str(x)):
        values = {name_generator(x): x for x in items}
        return NamedInts(**values)

    @classmethod
    def range(cls, from_value, to_value, name_generator=lambda x: str(x), step=1):
        values = {name_generator(x): x for x in range(from_value, to_value + 1, step)}
        return NamedInts(**values)

    def flag_names(self, value):
        unknown_bits = value
        for k in self._indexed:
            assert bin(k).count('1') == 1
            if k & value == k:
                unknown_bits &= ~k
                yield str(self._indexed[k])

        if unknown_bits:
            yield 'unknown:%06X' % unknown_bits

    def __getitem__(self, index):
        if isinstance(index, int):
            if index in self._indexed:
                return self._indexed[int(index)]
            if self._fallback and isinstance(index, int):
                value = NamedInt(index, self._fallback(index))
                self._indexed[index] = value
                self._values = sorted(self._values + [value])
                return value

        elif is_string(index):
            if index in self.__dict__:
                return self.__dict__[index]

        elif isinstance(index, slice):
            if index.start is None and index.stop is None:
                return self._values[:]

            v_start = int(self._values[0]) if index.start is None else int(index.start)
            v_stop = (self._values[-1] + 1) if index.stop is None else int(index.stop)

            if v_start > v_stop or v_start > self._values[-1] or v_stop <= self._values[0]:
                return []

            if v_start <= self._values[0] and v_stop > self._values[-1]:
                return self._values[:]

            start_index = 0
            stop_index = len(self._values)
            for i, value in enumerate(self._values):
                if value < v_start:
                    start_index = i + 1
                elif index.stop is None:
                    break
                if value >= v_stop:
                    stop_index = i
                    break

            return self._values[start_index:stop_index]

    def __setitem__(self, index, name):
        assert isinstance(index, int), type(index)
        if isinstance(name, NamedInt):
            assert int(index) == int(name), repr(index) + ' ' + repr(name)
            value = name
        elif is_string(name):
            value = NamedInt(index, name)
        else:
            raise TypeError('name must be a string')

        if str(value) in self.__dict__:
            raise ValueError('%s (%d) already known' % (value, int(value)))
        if int(value) in self._indexed:
            raise ValueError('%d (%s) already known' % (int(value), value))

        self._values = sorted(self._values + [value])
        self.__dict__[str(value)] = value
        self._indexed[int(value)] = value

    def __contains__(self, value):
        if isinstance(value, int):
            return value in self._indexed
        elif is_string(value):
            return value in self.__dict__

    def __iter__(self):
        for v in self._values:
            yield v

    def __len__(self):
        return len(self._values)

    def __repr__(self):
        return 'NamedInts(%s)' % ', '.join(repr(v) for v in self._values)


def strhex(x):
    assert x is not None
    """Produce a hex-string representation of a sequence of bytes."""
    return _hexlify(x).decode('ascii').upper()


def bytes2int(x):
    """Convert a bytes string to an int.
    The bytes are assumed to be in most-significant-first order.
    """
    assert isinstance(x, bytes)
    assert len(x) < 9
    qx = (b'\x00' * 8) + x
    result, = unpack('!Q', qx[-8:])
    # assert x == int2bytes(result, len(x))
    return result


def int2bytes(x, count=None):
    """Convert an int to a bytes representation.
    The bytes are ordered in most-significant-first order.
    If 'count' is not given, the necessary number of bytes is computed.
    """
    assert isinstance(x, int)
    result = pack('!Q', x)
    assert isinstance(result, bytes)
    # assert x == bytes2int(result)

    if count is None:
        return result.lstrip(b'\x00')

    assert isinstance(count, int)
    assert count > 0
    assert x.bit_length() <= count * 8
    return result[-count:]


class KwException(Exception):
    """An exception that remembers all arguments passed to the constructor.
    They can be later accessed by simple member access.
    """
    def __init__(self, **kwargs):
        super(KwException, self).__init__(kwargs)

    def __getattr__(self, k):
        try:
            return super(KwException, self).__getattr__(k)
        except AttributeError:
            return self.args[0][k]


"""Firmware information."""
FirmwareInfo = namedtuple('FirmwareInfo', ['kind', 'name', 'version', 'extras'])
"""Reprogrammable keys information."""
ReprogrammableKeyInfo = namedtuple('ReprogrammableKeyInfo', ['index', 'key', 'task', 'flags'])

ReprogrammableKeyInfoV4 = namedtuple(
    'ReprogrammableKeyInfoV4', ['index', 'key', 'task', 'flags', 'pos', 'group', 'group_mask', 'remapped']
)

del namedtuple



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Logitech Unifying Receiver API.

from __future__ import absolute_import, division, print_function, unicode_literals

from logging import DEBUG as _DEBUG
from logging import getLogger

from . import special_keys
from .common import FirmwareInfo as _FirmwareInfo
from .common import KwException as _KwException
from .common import NamedInts as _NamedInts
from .common import ReprogrammableKeyInfo as _ReprogrammableKeyInfo
from .common import ReprogrammableKeyInfoV4 as _ReprogrammableKeyInfoV4
from .common import pack as _pack
from .common import unpack as _unpack

_log = getLogger(__name__)
del getLogger

#
#
#

# <FeaturesSupported.xml sed '/LD_FID_/{s/.*LD_FID_/\t/;s/"[ \t]*Id="/=/;s/" \/>/,/p}' | sort -t= -k2
# additional features names taken from https://github.com/cvuchener/hidpp and
# https://github.com/Logitech/cpg-docs/tree/master/hidpp20
"""Possible features available on a Logitech device.

A particular device might not support all these features, and may support other
unknown features as well.
"""
FEATURE = _NamedInts(
    ROOT=0x0000,
    FEATURE_SET=0x0001,
    FEATURE_INFO=0x0002,
    # Common
    DEVICE_FW_VERSION=0x0003,
    DEVICE_UNIT_ID=0x0004,
    DEVICE_NAME=0x0005,
    DEVICE_GROUPS=0x0006,
    DEVICE_FRIENDLY_NAME=0x0007,
    KEEP_ALIVE=0x0008,
    RESET=0x0020,  # "Config Change"
    CRYPTO_ID=0x0021,
    TARGET_SOFTWARE=0x0030,
    WIRELESS_SIGNAL_STRENGTH=0x0080,
    DFUCONTROL_LEGACY=0x00C0,
    DFUCONTROL_UNSIGNED=0x00C1,
    DFUCONTROL_SIGNED=0x00C2,
    DFU=0x00D0,
    BATTERY_STATUS=0x1000,
    BATTERY_VOLTAGE=0x1001,
    CHARGING_CONTROL=0x1010,
    LED_CONTROL=0x1300,
    GENERIC_TEST=0x1800,
    DEVICE_RESET=0x1802,
    OOBSTATE=0x1805,
    CONFIG_DEVICE_PROPS=0x1806,
    CHANGE_HOST=0x1814,
    HOSTS_INFO=0x1815,
    BACKLIGHT=0x1981,
    BACKLIGHT2=0x1982,
    BACKLIGHT3=0x1983,
    PRESENTER_CONTROL=0x1A00,
    SENSOR_3D=0x1A01,
    REPROG_CONTROLS=0x1B00,
    REPROG_CONTROLS_V2=0x1B01,
    REPROG_CONTROLS_V2_2=0x1B02,  # LogiOptions 2.10.73 features.xml
    REPROG_CONTROLS_V3=0x1B03,
    REPROG_CONTROLS_V4=0x1B04,
    REPORT_HID_USAGE=0x1BC0,
    PERSISTENT_REMAPPABLE_ACTION=0x1C00,
    WIRELESS_DEVICE_STATUS=0x1D4B,
    REMAINING_PAIRING=0x1DF0,
    FIRMWARE_PROPERTIES=0x1F1F,
    ADC_MEASUREMENT=0x1F20,
    # Mouse
    LEFT_RIGHT_SWAP=0x2001,
    SWAP_BUTTON_CANCEL=0x2005,
    POINTER_AXIS_ORIENTATION=0x2006,
    VERTICAL_SCROLLING=0x2100,
    SMART_SHIFT=0x2110,
    HI_RES_SCROLLING=0x2120,
    HIRES_WHEEL=0x2121,
    LOWRES_WHEEL=0x2130,
    THUMB_WHEEL=0x2150,
    MOUSE_POINTER=0x2200,
    ADJUSTABLE_DPI=0x2201,
    POINTER_SPEED=0x2205,
    ANGLE_SNAPPING=0x2230,
    SURFACE_TUNING=0x2240,
    HYBRID_TRACKING=0x2400,
    # Keyboard
    FN_INVERSION=0x40A0,
    NEW_FN_INVERSION=0x40A2,
    K375S_FN_INVERSION=0x40A3,
    ENCRYPTION=0x4100,
    LOCK_KEY_STATE=0x4220,
    SOLAR_DASHBOARD=0x4301,
    KEYBOARD_LAYOUT=0x4520,
    KEYBOARD_DISABLE_KEYS=0x4521,
    KEYBOARD_DISABLE_BY_USAGE=0x4522,
    DUALPLATFORM=0x4530,
    MULTIPLATFORM=0x4531,
    KEYBOARD_LAYOUT_2=0x4540,
    CROWN=0x4600,
    # Touchpad
    TOUCHPAD_FW_ITEMS=0x6010,
    TOUCHPAD_SW_ITEMS=0x6011,
    TOUCHPAD_WIN8_FW_ITEMS=0x6012,
    TAP_ENABLE=0x6020,
    TAP_ENABLE_EXTENDED=0x6021,
    CURSOR_BALLISTIC=0x6030,
    TOUCHPAD_RESOLUTION=0x6040,
    TOUCHPAD_RAW_XY=0x6100,
    TOUCHMOUSE_RAW_POINTS=0x6110,
    TOUCHMOUSE_6120=0x6120,
    GESTURE=0x6500,
    GESTURE_2=0x6501,
    # Gaming Devices
    GKEY=0x8010,
    MKEYS=0x8020,
    MR=0x8030,
    BRIGHTNESS_CONTROL=0x8040,
    REPORT_RATE=0x8060,
    COLOR_LED_EFFECTS=0x8070,
    RGB_EFFECTS=0X8071,
    PER_KEY_LIGHTING=0x8080,
    PER_KEY_LIGHTING_V2=0x8081,
    MODE_STATUS=0x8090,
    ONBOARD_PROFILES=0x8100,
    MOUSE_BUTTON_SPY=0x8110,
    LATENCY_MONITORING=0x8111,
    GAMING_ATTACHMENTS=0x8120,
    FORCE_FEEDBACK=0x8123,
    SIDETONE=0x8300,
    EQUALIZER=0x8310,
    HEADSET_OUT=0x8320,
)
FEATURE._fallback = lambda x: 'unknown:%04X' % x

FEATURE_FLAG = _NamedInts(internal=0x20, hidden=0x40, obsolete=0x80)

DEVICE_KIND = _NamedInts(
    keyboard=0x00, remote_control=0x01, numpad=0x02, mouse=0x03, touchpad=0x04, trackball=0x05, presenter=0x06, receiver=0x07
)

FIRMWARE_KIND = _NamedInts(Firmware=0x00, Bootloader=0x01, Hardware=0x02, Other=0x03)

BATTERY_OK = lambda status: status not in (BATTERY_STATUS.invalid_battery, BATTERY_STATUS.thermal_error)

BATTERY_STATUS = _NamedInts(
    discharging=0x00,
    recharging=0x01,
    almost_full=0x02,
    full=0x03,
    slow_recharge=0x04,
    invalid_battery=0x05,
    thermal_error=0x06
)

CHARGE_STATUS = _NamedInts(charging=0x00, full=0x01, not_charging=0x02, error=0x07)

CHARGE_LEVEL = _NamedInts(average=50, full=90, critical=5)

CHARGE_TYPE = _NamedInts(standard=0x00, fast=0x01, slow=0x02)

ERROR = _NamedInts(
    unknown=0x01,
    invalid_argument=0x02,
    out_of_range=0x03,
    hardware_error=0x04,
    logitech_internal=0x05,
    invalid_feature_index=0x06,
    invalid_function=0x07,
    busy=0x08,
    unsupported=0x09
)

#
#
#


class FeatureNotSupported(_KwException):
    """Raised when trying to request a feature not supported by the device."""
    pass


class FeatureCallError(_KwException):
    """Raised if the device replied to a feature call with an error."""
    pass


#
#
#


class FeaturesArray(object):
    """A sequence of features supported by a HID++ 2.0 device."""
    __slots__ = ('supported', 'device', 'features', 'non_features')
    assert FEATURE.ROOT == 0x0000

    def __init__(self, device):
        assert device is not None
        self.device = device
        self.supported = True
        self.features = None
        self.non_features = set()

    def __del__(self):
        self.supported = False
        self.device = None
        self.features = None

    def _check(self):
        # print (self.device, "check", self.supported, self.features, self.device.protocol)
        if self.supported:
            assert self.device
            if self.features is not None:
                return True

            if not self.device.online:
                # device is not connected right now, will have to try later
                return False

            # I _think_ this is universally true
            if self.device.protocol and self.device.protocol < 2.0:
                self.supported = False
                self.device.features = None
                self.device = None
                return False

            reply = self.device.request(0x0000, _pack('!H', FEATURE.FEATURE_SET))
            if reply is None:
                self.supported = False
            else:
                fs_index = ord(reply[0:1])
                if fs_index:
                    count = self.device.request(fs_index << 8)
                    if count is None:
                        _log.warn('FEATURE_SET found, but failed to read features count')
                        # most likely the device is unavailable
                        return False
                    else:
                        count = ord(count[:1])
                        assert count >= fs_index
                        self.features = [None] * (1 + count)
                        self.features[0] = FEATURE.ROOT
                        self.features[fs_index] = FEATURE.FEATURE_SET
                        return True
                else:
                    self.supported = False

        return False

    __bool__ = __nonzero__ = _check

    def __getitem__(self, index):
        if self._check():
            if isinstance(index, int):
                if index < 0 or index >= len(self.features):
                    raise IndexError(index)

                if self.features[index] is None:
                    feature = self.device.feature_request(FEATURE.FEATURE_SET, 0x10, index)
                    if feature:
                        feature, = _unpack('!H', feature[:2])
                        self.features[index] = FEATURE[feature]

                return self.features[index]

            elif isinstance(index, slice):
                indices = index.indices(len(self.features))
                return [self.__getitem__(i) for i in range(*indices)]

    def __contains__(self, featureId):
        """Tests whether the list contains given Feature ID"""
        if self._check():
            ivalue = int(featureId)
            if ivalue in self.non_features:
                return False

            may_have = False
            for f in self.features:
                if f is None:
                    may_have = True
                elif ivalue == int(f):
                    return True

            if may_have:
                reply = self.device.request(0x0000, _pack('!H', ivalue))
                if reply:
                    index = ord(reply[0:1])
                    if index:
                        self.features[index] = FEATURE[ivalue]
                        return True
                    else:
                        self.non_features.add(ivalue)
                        return False

    def index(self, featureId):
        """Gets the Feature Index for a given Feature ID"""
        if self._check():
            may_have = False
            ivalue = int(featureId)
            for index, f in enumerate(self.features):
                if f is None:
                    may_have = True
                elif ivalue == int(f):
                    return index

            if may_have:
                reply = self.device.request(0x0000, _pack('!H', ivalue))
                if reply:
                    index = ord(reply[0:1])
                    self.features[index] = FEATURE[ivalue]
                    return index

        raise ValueError('%r not in list' % featureId)

    def __iter__(self):
        if self._check():
            yield FEATURE.ROOT
            index = 1
            last_index = len(self.features)
            while index < last_index:
                yield self.__getitem__(index)
                index += 1

    def __len__(self):
        return len(self.features) if self._check() else 0


#
#
#


class KeysArray(object):
    """A sequence of key mappings supported by a HID++ 2.0 device."""
    __slots__ = ('device', 'keys', 'keyversion')

    def __init__(self, device, count):
        assert device is not None
        self.device = device
        self.keyversion = 0
        self.keys = [None] * count

    def __getitem__(self, index):
        if isinstance(index, int):
            if index < 0 or index >= len(self.keys):
                raise IndexError(index)

            # TODO: add here additional variants for other REPROG_CONTROLS
            if self.keys[index] is None:
                keydata = feature_request(self.device, FEATURE.REPROG_CONTROLS, 0x10, index)
                self.keyversion = 1
                if keydata is None:
                    keydata = feature_request(self.device, FEATURE.REPROG_CONTROLS_V4, 0x10, index)
                    self.keyversion = 4
                if keydata:
                    key, key_task, flags, pos, group, gmask = _unpack('!HHBBBB', keydata[:8])
                    ctrl_id_text = special_keys.CONTROL[key]
                    ctrl_task_text = special_keys.TASK[key_task]
                    if self.keyversion == 1:
                        self.keys[index] = _ReprogrammableKeyInfo(index, ctrl_id_text, ctrl_task_text, flags)
                    if self.keyversion == 4:
                        try:
                            mapped_data = feature_request(
                                self.device, FEATURE.REPROG_CONTROLS_V4, 0x20, key & 0xff00, key & 0xff
                            )
                            if mapped_data:
                                remap_key, remap_flag, remapped = _unpack('!HBH', mapped_data[:5])
                                # if key not mapped map it to itself for display
                                if remapped == 0:
                                    remapped = key
                        except Exception:
                            remapped = key
                            # remap_key = key
                            # remap_flag = 0

                        remapped_text = special_keys.CONTROL[remapped]
                        self.keys[index] = _ReprogrammableKeyInfoV4(
                            index, ctrl_id_text, ctrl_task_text, flags, pos, group, gmask, remapped_text
                        )

            return self.keys[index]

        elif isinstance(index, slice):
            indices = index.indices(len(self.keys))
            return [self.__getitem__(i) for i in range(*indices)]

    def index(self, value):
        for index, k in enumerate(self.keys):
            if k is not None and int(value) == int(k.key):
                return index

        for index, k in enumerate(self.keys):
            if k is None:
                k = self.__getitem__(index)
                if k is not None:
                    return index

    def __iter__(self):
        for k in range(0, len(self.keys)):
            yield self.__getitem__(k)

    def __len__(self):
        return len(self.keys)


#
#
#


def feature_request(device, feature, function=0x00, *params, no_reply=False):
    if device.online and device.features:
        if feature in device.features:
            feature_index = device.features.index(int(feature))
            return device.request((feature_index << 8) + (function & 0xFF), *params, no_reply=no_reply)


def get_firmware(device):
    """Reads a device's firmware info.

    :returns: a list of FirmwareInfo tuples, ordered by firmware layer.
    """
    count = feature_request(device, FEATURE.DEVICE_FW_VERSION)
    if count:
        count = ord(count[:1])

        fw = []
        for index in range(0, count):
            fw_info = feature_request(device, FEATURE.DEVICE_FW_VERSION, 0x10, index)
            if fw_info:
                level = ord(fw_info[:1]) & 0x0F
                if level == 0 or level == 1:
                    name, version_major, version_minor, build = _unpack('!3sBBH', fw_info[1:8])
                    version = '%02X.%02X' % (version_major, version_minor)
                    if build:
                        version += '.B%04X' % build
                    extras = fw_info[9:].rstrip(b'\x00') or None
                    fw_info = _FirmwareInfo(FIRMWARE_KIND[level], name.decode('ascii'), version, extras)
                elif level == FIRMWARE_KIND.Hardware:
                    fw_info = _FirmwareInfo(FIRMWARE_KIND.Hardware, '', str(ord(fw_info[1:2])), None)
                else:
                    fw_info = _FirmwareInfo(FIRMWARE_KIND.Other, '', '', None)

                fw.append(fw_info)
                # if _log.isEnabledFor(_DEBUG):
                #     _log.debug("device %d firmware %s", devnumber, fw_info)
        return tuple(fw)


def get_kind(device):
    """Reads a device's type.

    :see DEVICE_KIND:
    :returns: a string describing the device type, or ``None`` if the device is
    not available or does not support the ``DEVICE_NAME`` feature.
    """
    kind = feature_request(device, FEATURE.DEVICE_NAME, 0x20)
    if kind:
        kind = ord(kind[:1])
        # if _log.isEnabledFor(_DEBUG):
        #     _log.debug("device %d type %d = %s", devnumber, kind, DEVICE_KIND[kind])
        return DEVICE_KIND[kind]


def get_name(device):
    """Reads a device's name.

    :returns: a string with the device name, or ``None`` if the device is not
    available or does not support the ``DEVICE_NAME`` feature.
    """
    name_length = feature_request(device, FEATURE.DEVICE_NAME)
    if name_length:
        name_length = ord(name_length[:1])

        name = b''
        while len(name) < name_length:
            fragment = feature_request(device, FEATURE.DEVICE_NAME, 0x10, len(name))
            if fragment:
                name += fragment[:name_length - len(name)]
            else:
                _log.error('failed to read whole name of %s (expected %d chars)', device, name_length)
                return None

        return name.decode('ascii')


def get_battery(device):
    """Reads a device's battery level."""
    battery = feature_request(device, FEATURE.BATTERY_STATUS)
    if battery:
        discharge, dischargeNext, status = _unpack('!BBB', battery[:3])
        discharge = None if discharge == 0 else discharge
        if _log.isEnabledFor(_DEBUG):
            _log.debug(
                'device %d battery %d%% charged, next level %d%% charge, status %d = %s', device.number, discharge,
                dischargeNext, status, BATTERY_STATUS[status]
            )
        return discharge, BATTERY_STATUS[status], dischargeNext


def get_voltage(device):
    battery_voltage = feature_request(device, FEATURE.BATTERY_VOLTAGE)
    if battery_voltage:
        return decipher_voltage(battery_voltage)


# modified to be much closer to battery reports
def decipher_voltage(voltage_report):
    voltage, flags = _unpack('>HB', voltage_report[:3])
    status = BATTERY_STATUS.discharging
    charge_sts = ERROR.unknown
    charge_lvl = CHARGE_LEVEL.average
    charge_type = CHARGE_TYPE.standard

    if flags & (1 << 7):
        status = BATTERY_STATUS.recharging
        charge_sts = CHARGE_STATUS[flags & 0x03]
    if charge_sts is None:
        charge_sts = ERROR.unknown
    elif charge_sts == CHARGE_STATUS.full:
        charge_lvl = CHARGE_LEVEL.full
        status = BATTERY_STATUS.full
    if (flags & (1 << 3)):
        charge_type = CHARGE_TYPE.fast
    elif (flags & (1 << 4)):
        charge_type = CHARGE_TYPE.slow
        status = BATTERY_STATUS.slow_recharge
    elif (flags & (1 << 5)):
        charge_lvl = CHARGE_LEVEL.critical

    if _log.isEnabledFor(_DEBUG):
        _log.debug(
            'device ???, battery voltage %d mV, charging = %s, charge status %d = %s, charge level %s, charge type %s',
            voltage, status, (flags & 0x03), charge_sts, charge_lvl, charge_type
        )

    return charge_lvl, status, voltage, charge_sts, charge_type


def get_keys(device):
    # TODO: add here additional variants for other REPROG_CONTROLS
    count = feature_request(device, FEATURE.REPROG_CONTROLS)
    if count is None:
        count = feature_request(device, FEATURE.REPROG_CONTROLS_V4)
    if count:
        return KeysArray(device, ord(count[:1]))


def get_mouse_pointer_info(device):
    pointer_info = feature_request(device, FEATURE.MOUSE_POINTER)
    if pointer_info:
        dpi, flags = _unpack('!HB', pointer_info[:3])
        acceleration = ('none', 'low', 'med', 'high')[flags & 0x3]
        suggest_os_ballistics = (flags & 0x04) != 0
        suggest_vertical_orientation = (flags & 0x08) != 0
        return {
            'dpi': dpi,
            'acceleration': acceleration,
            'suggest_os_ballistics': suggest_os_ballistics,
            'suggest_vertical_orientation': suggest_vertical_orientation
        }


def get_vertical_scrolling_info(device):
    vertical_scrolling_info = feature_request(device, FEATURE.VERTICAL_SCROLLING)
    if vertical_scrolling_info:
        roller, ratchet, lines = _unpack('!BBB', vertical_scrolling_info[:3])
        roller_type = (
            'reserved', 'standard', 'reserved', '3G', 'micro', 'normal touch pad', 'inverted touch pad', 'reserved'
        )[roller]
        return {'roller': roller_type, 'ratchet': ratchet, 'lines': lines}


def get_hi_res_scrolling_info(device):
    hi_res_scrolling_info = feature_request(device, FEATURE.HI_RES_SCROLLING)
    if hi_res_scrolling_info:
        mode, resolution = _unpack('!BB', hi_res_scrolling_info[:2])
        return mode, resolution


def get_pointer_speed_info(device):
    pointer_speed_info = feature_request(device, FEATURE.POINTER_SPEED)
    if pointer_speed_info:
        pointer_speed_hi, pointer_speed_lo = _unpack('!BB', pointer_speed_info[:2])
        # if pointer_speed_lo > 0:
        #     pointer_speed_lo = pointer_speed_lo
        return pointer_speed_hi + pointer_speed_lo / 256


def get_lowres_wheel_status(device):
    lowres_wheel_status = feature_request(device, FEATURE.LOWRES_WHEEL)
    if lowres_wheel_status:
        wheel_flag = _unpack('!B', lowres_wheel_status[:1])[0]
        wheel_reporting = ('HID', 'HID++')[wheel_flag & 0x01]
        return wheel_reporting


def get_hires_wheel(device):
    caps = feature_request(device, FEATURE.HIRES_WHEEL, 0x00)
    mode = feature_request(device, FEATURE.HIRES_WHEEL, 0x10)
    ratchet = feature_request(device, FEATURE.HIRES_WHEEL, 0x030)

    if caps and mode and ratchet:
        # Parse caps
        multi, flags = _unpack('!BB', caps[:2])

        has_invert = (flags & 0x08) != 0
        has_ratchet = (flags & 0x04) != 0

        # Parse mode
        wheel_mode, reserved = _unpack('!BB', mode[:2])

        target = (wheel_mode & 0x01) != 0
        res = (wheel_mode & 0x02) != 0
        inv = (wheel_mode & 0x04) != 0

        # Parse Ratchet switch
        ratchet_mode, reserved = _unpack('!BB', ratchet[:2])

        ratchet = (ratchet_mode & 0x01) != 0

        return multi, has_invert, has_ratchet, inv, res, target, ratchet


def get_new_fn_inversion(device):
    state = feature_request(device, FEATURE.NEW_FN_INVERSION, 0x00)
    if state:
        inverted, default_inverted = _unpack('!BB', state[:2])
        inverted = (inverted & 0x01) != 0
        default_inverted = (default_inverted & 0x01) != 0
        return inverted, default_inverted


def get_host_names(device):
    state = feature_request(device, FEATURE.HOSTS_INFO, 0x00)
    host_names = {}
    if state:
        _ignore, _ignore, numHosts, currentHost = _unpack('!BBBB', state[:4])
        for host in range(0, numHosts):
            hostinfo = feature_request(device, FEATURE.HOSTS_INFO, 0x10, host)
            _ignore, status, _ignore, numPages, nameLen, _ignore = _unpack('!BBBBBB', hostinfo[:6])
            name = ''
            remaining = nameLen
            while remaining > 0:
                name_piece = feature_request(device, FEATURE.HOSTS_INFO, 0x30, host, nameLen - remaining)
                name += name_piece[2:2 + min(remaining, 14)].decode()
                remaining = max(0, remaining - 14)
            host_names[host] = (bool(status), name)
    return host_names


def set_host_name(device, name):
    state = feature_request(device, FEATURE.HOSTS_INFO, 0x00)
    if state:
        flags = _unpack('!B', state[:1])[0]
        if flags & 0x02:
            hn = name[:min(14, name.find('.'))] if name.find('.') >= 0 else name
            response = feature_request(device, FEATURE.HOSTS_INFO, 0x40, 0xff, 0, hn)
            return response


def get_polling_rate(device):
    state = feature_request(device, FEATURE.REPORT_RATE, 0x10)
    if state:
        rate = _unpack('!B', state[:1])[0]
        return rate
